template <typename word, typename resizable_array>
resizable_bitarray<word, resizable_array>::resizable_bitarray (
): N(0), B() {
}

template <typename word, typename resizable_array>
template <typename integer>
resizable_bitarray<word, resizable_array>::resizable_bitarray (
  integer number_of_bits, 
  bool initial_value,
  const resizable_array&
): N(number_of_bits), B() {
  typedef typename resizable_array::iterator position;
  (*this).B.resize((number_of_bits >> (*this).log_of_w) + 1);
  word v = (initial_value) ? ~word(0) : word(0);
  for (position p = (*this).B.begin(); p != (*this).B.end(); ++p) {
    *p = v;
  }
}
 
template <typename word, typename resizable_array>
cphstl::resizable_bitarray<word, resizable_array>::resizable_bitarray (
  const cphstl::resizable_bitarray<word, resizable_array>& C
): N(C.N), B(C.B) {
}
   
template <typename word, typename resizable_array>
resizable_bitarray<word, resizable_array>::~resizable_bitarray (
) {
}
  
template <typename word, typename resizable_array>
typename resizable_bitarray<word, resizable_array>::size_type
resizable_bitarray<word, resizable_array>::size() const {
  return (*this).N;
}

template <typename word, typename resizable_array>
bool 
resizable_bitarray<word, resizable_array>::empty (
) const {
  return (*this).N == 0;
}

template <typename word, typename resizable_array>
typename resizable_bitarray<word, resizable_array>::size_type
resizable_bitarray<word, resizable_array>::capacity (
) const {
  return (*this).B.capacity() << (*this).log_of_w;
}

template <typename word, typename resizable_array>
template <typename integer>
void 
resizable_bitarray<word, resizable_array>::reserve (
  integer n
) {
  (*this).B.reserve((n >> (*this).log_of_w) + 1);
}
  
template <typename word, typename resizable_array>
template <typename index>
void 
resizable_bitarray<word, resizable_array>::set (
  index I,
  bool x
) {
  index i = I >> (*this).log_of_w;
  index lambda = I & (*this).tail_mask;
  word v = 1 << lambda;
  B[i] = (x) ? B[i] | v : B[i] & ~v;
}

template <typename word, typename resizable_array>
template <typename index>
bool 
resizable_bitarray<word, resizable_array>::get (
  index I
) const {
  index i = I >> (*this).log_of_w;
  index lambda = I & (*this).tail_mask;
  word v = 1 << lambda;
  v = B[i] & v;
  
  // Alternativ
  //return (v >> lambda);

  return (v > 0);
}

template <typename word, typename resizable_array>
template <typename index, typename length>
void 
resizable_bitarray<word, resizable_array>::set (
  index I,
  length L,
  word set_w
) {
  index i = I >> (*this).log_of_w;
  index lambda = I & (*this).tail_mask;
  length first_length = (*this).w - lambda;
  word one_mask = (1 << L) - 1;
  if (L > first_length) {
    word clear_mask = ~(one_mask << lambda);
    B[i] = (B[i] & clear_mask) | (set_w << lambda);
    clear_mask = ~(one_mask >> first_length);
    B[i + 1] = (B[i + 1] & clear_mask) | (set_w >> first_length);    
    return;
  }
  word clear_mask = ~(one_mask << lambda);
  B[i] = (B[i] & clear_mask) | (set_w << lambda) ;
}


template <typename word, typename resizable_array>
template <typename index, typename length>
word 
resizable_bitarray<word, resizable_array>::get (
  index I,
  length L
) const {
  index i = I >> (*this).log_of_w;
  index lambda = I & (*this).tail_mask;
  length first_length = (*this).w - lambda;
  word one_mask = (1 << L) - 1;
  if (L <= first_length) {
    return ((B[i] >> lambda) & one_mask);
  }
  one_mask = one_mask >> first_length;
  return ((B[i + 1] & one_mask) << first_length) | (B[i] >> lambda);
}

template <typename word, typename resizable_array>
void 
resizable_bitarray<word, resizable_array>::push_back (
  bool x
){
  if ((*this).B.size() << (*this).log_of_w <= (*this).N) {
    (*this).B.push_back(word());
  }
  (*this).set((*this).N, x);    
  ++(*this).N;
}

template <typename word, typename resizable_array>
void 
resizable_bitarray<word, resizable_array>::pop_back() {
  if (((*this).B.size() - ((*this).N >> (*this).log_of_w)) > 2) {
    (*this).B.pop_back();
  }
  --(*this).N;
}

