template <typename word, typename resizable_array>
resizable_bitarray<word, resizable_array>::resizable_bitarray (
): N(0), RA() {
}

template <typename word, typename resizable_array>
template <typename integer>
resizable_bitarray<word, resizable_array>::resizable_bitarray (
  integer number_of_bits, 
  bool initial_value,
  const resizable_array&
): N(number_of_bits), RA() {
  typedef typename resizable_array::iterator position;
  (*this).RA.resize((number_of_bits >> (*this).log_of_w) + 1);
  word v = (initial_value) ? ~word(0) : word(0);
  for (position p = (*this).RA.begin(); p != (*this).RA.end(); ++p) {
    *p = v;
  }
}
 
template <typename word, typename resizable_array>
cphstl::resizable_bitarray<word, resizable_array>::resizable_bitarray (
  const cphstl::resizable_bitarray<word, resizable_array>& C
): N(C.N), RA(C.RA) {
}
   
template <typename word, typename resizable_array>
resizable_bitarray<word, resizable_array>::~resizable_bitarray (
) {
}
  
template <typename word, typename resizable_array>
typename resizable_bitarray<word, resizable_array>::size_type
resizable_bitarray<word, resizable_array>::size() const {
  return (*this).N;
}

template <typename word, typename resizable_array>
bool 
resizable_bitarray<word, resizable_array>::empty (
) const {
  return (*this).N == 0;
}

template <typename word, typename resizable_array>
typename resizable_bitarray<word, resizable_array>::size_type
resizable_bitarray<word, resizable_array>::capacity (
) const {
  return (*this).RA.capacity() << (*this).log_of_w;
}

template <typename word, typename resizable_array>
template <typename integer>
void 
resizable_bitarray<word, resizable_array>::reserve (
  integer n
) {
  (*this).RA.reserve((n >> (*this).log_of_w) + 1);
}
  
template <typename word, typename resizable_array>
template <typename index>
void 
resizable_bitarray<word, resizable_array>::set (
  index I,
  bool x
) {
  index i = I >> (*this).log_of_w;
  index lambda = I & (*this).tail_mask;
  word v = 1 << lambda;
  RA[i] = (x) ? RA[i] | v : RA[i] & ~v;
}

template <typename word, typename resizable_array>
template <typename index>
bool 
resizable_bitarray<word, resizable_array>::get (
  index I
) const {
  index i = I >> (*this).log_of_w;
  index lambda = I & (*this).tail_mask;
  word v = 1 << lambda;
  v = RA[i] & v;
  
  // Alternativ
  //return (v >> lambda);

  return (v > 0);
}

template <typename word, typename resizable_array>
template <typename index, typename length>
void 
resizable_bitarray<word, resizable_array>::set (
  index I,
  length L,
  word set_w
) {
  index i = I >> (*this).log_of_w;
  index lambda = I & (*this).tail_mask;
  length bottom_length = w - lambda;
  if (L <= bottom_length) bottom_length = L;
  word bottom_mask = (1 << bottom_length + 1) - 1;
  word v = set_w & bottom_mask;
  RA[i] = RA[i] + (v << lambda);
  if (L > length(w - lambda)) {
    RA[i + 1] = RA[i + 1] + set_w >> w - lambda;
  }
}

template <typename word, typename resizable_array>
template <typename index, typename length>
word 
resizable_bitarray<word, resizable_array>::get (
  index I,
  length L
) const {
  index i = I >> (*this).log_of_w;
  index lambda = I & (*this).tail_mask;
  length bottom_length = w - lambda;
  if (L <= bottom_length) {
    word bottom_mask = (1 << L) - 1 << lambda;
    word bv = (RA[i] & bottom_mask) >> lambda;
    return bv;
  }
  word bottom_mask = (1 << bottom_length) - 1 << lambda;
  word bv = (RA[i] & bottom_mask) >> lambda;
  word top_mask = (1 << L - bottom_length) - 1;
  word tv = RA[i + 1] & top_mask;
  return (tv << bottom_length) + bv;
}

template <typename word, typename resizable_array>
void 
resizable_bitarray<word, resizable_array>::push_back (
  bool x
){
  if ((*this).RA.size() << (*this).log_of_w <= (*this).N) {
    (*this).RA.push_back();
  }
  (*this).set((*this).N, x);    
  ++(*this).N;
}

template <typename word, typename resizable_array>
void 
resizable_bitarray<word, resizable_array>::pop_back() {
  if (((*this).RA.size() - ((*this).N >> (*this).log_of_w)) > 2) {
    (*this).RA.pop_back();
  }
  --(*this).N;
}

