#ifndef __cphstl_resizable_bitarray__
#define __cphstl_resizable_bitarray__

#include <vector> // defines std::vector
#include <cstddef> // defines std::sizeof() and std::size_t
#include <climits> // defines CHAR_BIT
#include <boost/integer/static_log2.hpp> // defines boost::static_log2

namespace cphstl {

  template <
    typename word = unsigned long, // must be an unsigned integer type
    typename resizable_array = std::vector<word>
  >
  class resizable_bitarray {
  public:
    typedef bool value_type;
    typedef std::size_t size_type;
    typedef word block_type;
    typedef resizable_array container_type;

    enum word_size { w = CHAR_BIT * sizeof(word) }; // w must be a power of 2
  
    explicit resizable_bitarray();

    template <typename integer>
    explicit resizable_bitarray(integer number_of_bits = 0, 
				bool inital_value = false,
				const resizable_array& = resizable_array());
    resizable_bitarray(const resizable_bitarray&);  
    ~resizable_bitarray();
  
    size_type size() const;
    bool empty() const;
    size_type capacity() const;
    template <typename integer>
    void reserve(integer);
    
    template <typename index>
    void set(index, bool);
    template <typename index>
    bool get(index) const;

    template <typename index, typename length>
    void set(index, length, word);
    template <typename index, typename length>
    word get(index, length) const;

    void push_back(bool);
    void pop_back();

    resizable_array RA;
  
  private:
    enum { log_of_w = boost::static_log2<w>::value };
    enum { tail_mask = w - 1 };
    size_type N;
  };

#include <resizable_bitarray.c++> // defines the member functions

} // namespace cphstl

#endif // __cphstl_resizable_bitarray__


/* 
Possible extensions:
- clear()
- resize()
- assign()
- operator=()
- max_size()
- operator==()
- operator<()
*/
