#include <iostream> // defines standard streams
#include <cassert> // defines assert
#include "input_generation.c++" // defines data generation routines

#include "binary_heap_ref.cpp"

int main () {
  typedef std::size_t element;
  typedef std::less<element> ordering;
  typedef std::vector<element> Container;
  const std::size_t bign = 100;
  const std::size_t repetitions = 100;
  element a[bign];
  Container C(bign);

  std::cout << "Testing interface" << std::endl;
  // OBS OBS N (C.size) present beclause of compile error
  random_sequence(C.begin(), C.end());
  binary_heap_ref<element, Container, ordering>
    bhr_first_random(C.size(), ordering(), C);
  bhr_first_random.Sort_binary_heap_ref(ordering(), C);

  random_sequence(a, a + bign);
  Container Z(bign, bign);
  binary_heap_ref<element, Container, ordering>
    bhr_third_random(a, a + bign, ordering(), Z);
  Container E(a, a + bign);
  Z.insert(Z.end(), E.begin(), E.end());
  bhr_third_random.Sort_binary_heap_ref(ordering(), Z);
  std::cout << "Interface ok" << std::endl;

  std::cout << "Testing pop binary_heap_ref, make_binary_heap_ref" << std::endl;
  for (std::size_t n = 0; n <= bign; ++n) {
    decreasing_sequence(a, a + n);
    Container F(0,0);
    binary_heap_ref<element, Container, ordering>
      bhr_decreasing(a, a + n, ordering(), F);
    Container FF(a, a + n);
    bhr_decreasing.Sort_binary_heap_ref(ordering(), FF);
  
    increasing_sequence(a, a + n);
    Container G(0, 0);
    binary_heap_ref<element, Container, ordering>
      bhr_increasing(a, a + n, ordering(), G);
    Container GG(a, a + n);
    bhr_increasing.Sort_binary_heap_ref(ordering(), GG);
    
    zero_sequence(a, a + n);
    Container H(0, 0);
    binary_heap_ref<element, Container, ordering>
      bhr_zero(a, a + n, ordering(), H);
    Container HH(a, a + n);
    bhr_zero.Sort_binary_heap_ref(ordering(), HH);
    
    binary_sequence(a, a + n);
    Container I(0, 0);
    binary_heap_ref<element, Container, ordering>
      bhr_binary(a, a + n, ordering(), I);
    Container II(a, a + n);
    bhr_binary.Sort_binary_heap_ref(ordering(), II);

    for (std::size_t i = 0; i < repetitions; ++i) {
      Container J(0, 0);
      random_sequence(a, a + n);
      binary_heap_ref<element, Container, ordering>
	bhr_random(a, a + n, ordering(), J);
      Container JJ(a, a + n);
      bhr_random.Sort_binary_heap_ref(ordering(), JJ);
    }
    std::cout << n << " ok" << std::endl;
  }

  std::cout << "Testing push binary_heap_ref" << std::endl;
  typedef void routine(element*, element*);
  routine* generate_sequence[] = {
    increasing_sequence,
    decreasing_sequence, 
    zero_sequence,
    random_sequence, 
    binary_sequence
  };  
  for (std::size_t n = 1; n <= bign; ++n) {
    for (unsigned int i = 0; i < sizeof(generate_sequence) / sizeof(routine*); ++i) {
      (void) generate_sequence[i](a, a + n);
      
      Container K(0, 0);
      binary_heap_ref<element, Container, ordering>
	bhr(a, a + 0, ordering(), K);
      
      for (unsigned int k = 1; k <= n; ++k) {
	bhr.push(a[k - 1]);
      }
      Container KK(a, a + n);
      bhr.Sort_binary_heap_ref(ordering(), KK);
    }
    std::cout << n << " ok" << std::endl;
  }
  return 0;
}
