#include <cstdlib> // defines std::size_t
#include <iterator> // defines std::iterator_traits
#include <vector> // defines std::vector
#include <functional> // defines std::less


template <class T, class Container = std::vector<T>, 
	  class Compare = std::less<typename Container::value_type> >
class navigation_pile {
 public:
  typedef typename Container::value_type value_type;
  typedef typename Container::size_type size_type;
  typedef Container container_type;

  // n: # of elements; N: capacity
  explicit navigation_pile(size_type N, 
			   const Compare& x = Compare(), 
			   const Container& Y = Container());
  template <class InputIterator>
  navigation_pile(InputIterator first, 
		  InputIterator last, 
		  size_type N = last - first,
		  const Compare& x = Compare(),
		  const Container& Y = Container() );
  ~navigation_pile();

  const  value_type& top() const;
  void push(const value_type& x);
  void pop();
  bool empty() const;
  size_type size() const;
  void sort_pile(size_type n, const Compare& x); // Made public for testing
  
 private:
  typedef typename Container::reference reference;
  typedef typename Container::const_reference const_reference;  
  typedef T element;
  typedef size_type element_index;
  typedef size_type node_index;
  typedef size_type branch_index;
  typedef size_type leaf_index;
  typedef size_type level;
  Container A; 
  std::vector<element_index> B;
  Compare less;
  size_type _n;
  size_type _N;
  level _eta;
  leaf_index _first_leaf;
  node_index _root;
  size_type _root_n;
  level depth(node_index i) const;
  level height(node_index i) const;
  node_index first_child(node_index i) const;
  node_index second_child(node_index i) const;
  node_index parent(node_index i) const;
  node_index ancestor(node_index i, level Delta) const;
  node_index root() const;
  leaf_index last_leaf() const;
  bool  is_first_child(node_index i) const;
  bool is_root(node_index i) const;
  leaf_index element_to_leaf(element_index ell) const;
  element_index leaf_to_element(leaf_index L) const;
  element_index jump_to_element(branch_index i) const;
  void make_pile();
  void handle_height_one_branch(branch_index i);
  node_index binary_search_on_special_path(node_index j, const element& x);
  void  update_partial_path(node_index j, node_index i, leaf_index K);
  node_index update_partial_path(leaf_index K);  
  node_index first_ancestor_with_two_children(leaf_index L);
  void first_two_ancestors_with_more_than_one_child(branch_index &i_1 , branch_index &i_2  , leaf_index L);
  void update_full_path(leaf_index L);
  void update_partial_path(branch_index j, branch_index i, leaf_index K, leaf_index L);  
  void insert_new_element(const element& z);
  void erase_last_element();
  node_index bottom_up_on_special_path(node_index j, const element& x);
  node_index sibling(node_index i) const;
  node_index pow_two(level i) const;
  int ilogb_zero(double i) const;
  size_type ceiling_log(size_type i) const;
};
