#include <iostream> // defines standard streams
#include <cassert> // defines assert
#include <cstdlib> // defines std:.size_t
#include <functional> // defines std::less
#include <algorithm> // defines std::copy
#include "input_generation.c++" // defines data generation routines

#include "navigation_pile.cpp"

int main () {
  typedef std::size_t element;
  typedef element* position;
  typedef std::less<element> ordering;
  typedef std::vector<element> Container;

  const std::size_t bign = 100;
  const std::size_t repetitions = 100;
  element a[bign];

  std::cout << "Testing interface" << std::endl;
  for (std::size_t n = 0; n <= bign; ++n) {
    Container C(n);

    random_sequence(C.begin(), C.end());
    navigation_pile<element, Container, ordering>
      np_first_interface_random(C.size(), ordering(), C);
    np_first_interface_random.sort_pile(C.size(), ordering());

    random_sequence(a, a + bign);
    navigation_pile<element, Container, ordering>
      np_second_random(a, a + bign, bign, ordering());
    np_second_random.sort_pile(bign, ordering());

    std::cout << "Interface ok for n " << n << std::endl;
  }

  std::cout << "Testing pop, make_pile" << std::endl;

  for (std::size_t n = 0; n <= bign; ++n) {
    
    decreasing_sequence(a, a + n);
    navigation_pile<element, Container, ordering>
      np_decreasing(a, a + n, n, ordering());
    np_decreasing.sort_pile(n, ordering());
    
    increasing_sequence(a, a + n);
    navigation_pile<element, Container, ordering>
      np_increasing(a, a + n, n, ordering());
    np_increasing.sort_pile(n, ordering());
    
    zero_sequence(a, a + n);
    navigation_pile<element, Container, ordering>
      np_zero(a, a + n, n, ordering());
    np_zero.sort_pile(n, ordering());
    
    binary_sequence(a, a + n);
    navigation_pile<element, Container, ordering>
      np_binary(a, a + n, n, ordering());
    np_binary.sort_pile(n, ordering());
    
    for (std::size_t i = 0; i < repetitions; ++i) {
      random_sequence(a, a + n);
      navigation_pile<element, Container, ordering>
	np_random(a, a + n, n, ordering());
      np_random.sort_pile(n, ordering());
    }
    std::cout << n << " ok" << std::endl;
  }

  std::cout << "Testing push" << std::endl;

  typedef void routine(position, position);
  routine* generate_sequence[] = {
    increasing_sequence,
    decreasing_sequence, 
    zero_sequence,
    random_sequence, 
    binary_sequence
  };

  for (std::size_t n = 1; n <= bign; ++n) {
    for (unsigned int i = 0; i < sizeof(generate_sequence) / sizeof(routine*); ++i) {
      (void) generate_sequence[i](a, a + n);
      
      navigation_pile<element, Container, ordering>
	np(a, a + 0, n, ordering());
      
      for (unsigned int k = 1; k <= n; ++k) {
	np.push(a[k - 1]);
      }
      np.sort_pile(n, ordering());
    }
    std::cout << n << " ok" << std::endl;
  }
  return 0;
}
