import benchmark
import os

class stdsort_case(benchmark.case):
  def __init__(self, n, k, element, function, directory):
    benchmark.case.__init__(self)
    self.n = n
#    self.computer = 'pc-014.diku.dk'
#    self.compiler = '/scratch/g++3.0.4/bin/g++'
#    self.compiler = '/usr/bin/g++-3.2' 
    self.compiler = 'g++-3' 
    self.compiler_options.extend(['-O6'])
    home = os.environ['HOME']
    self.include_paths.extend([
      home + '/CPHSTL/Report/In-place-multiway-heaps/Program/Common/',
      home + '/CPHSTL/Report/In-place-multiway-heaps/Program/' + directory + '/' 
      ])
    self.include_files.extend(['<vector>', '<functional>', function + '.c++'])
    self.dual_exists = 1
    self.constructor_call = \
      'experiment<' + \
         element + ',' + \
         'std::less<' + element + '>,' +\
        'std::vector<' + element + '>,' +\
        'std::vector<' + element + '>::iterator' +\
      '>(' + str(n) + ', ' + str(k) + ')'
    self.time_unit = 'ns'
    self.driver_file = self.generate_cpu_time_driver()

  def output(self):
    return (self.n, float(self.driver_output) / self.n)


class cphstlsort_case(benchmark.case):
  def __init__(self, n, k, element, heap, function, directory):
    benchmark.case.__init__(self)
    self.n = n
#    self.computer = 'pc-014.diku.dk'
#    self.compiler = '/scratch/g++3.0.4/bin/g++'
#    self.compiler = '/usr/bin/g++-3.2' 
    self.compiler = 'g++-3' 
    self.compiler_options.extend(['-O6'])
    home = os.environ['HOME']
    self.include_paths.extend([
      home + '/CPHSTL/Report/In-place-multiway-heaps/Program/Common/',
      home + '/CPHSTL/Report/In-place-multiway-heaps/Program/' + directory + '/' 
      ])
    self.include_files.extend([heap + '.c++', '<vector>', '<functional>', function + '.c++'])
    self.dual_exists = 1
    self.constructor_call = \
      'experiment<' + \
         element + ',' + \
         'std::less<' + element + '>,' +\
        'std::vector<' + element + '>,' +\
        'std::vector<' + element + '>::iterator' +\
      '>(' + str(n) + ', ' + str(k) + ')'
    self.time_unit = 'ns'
    self.driver_file = self.generate_cpu_time_driver()

  def output(self):
    return (self.n, float(self.driver_output) / self.n)

class sandersort_case(benchmark.case):
  def __init__(self, n, k, element, function, directory):
    benchmark.case.__init__(self)
    self.n = n
#    self.computer = 'pc-014.diku.dk'
#    self.compiler = '/scratch/g++3.0.4/bin/g++'
#    self.compiler = '/usr/bin/g++-3.2' 
    self.compiler = 'g++-3' 
    self.compiler_options.extend(['-O6'])
    home = os.environ['HOME']
    self.include_paths.extend([
      home + '/CPHSTL/Report/In-place-multiway-heaps/Program/Common/',
      home + '/CPHSTL/Report/In-place-multiway-heaps/Program/' + directory + '/' 
      ])
    self.include_files.extend(['heap2.h', '<vector>', '<functional>', function + '.c++'])
    self.dual_exists = 1
    self.constructor_call = \
      'experiment<' + \
         element + ',' + \
         'std::less<' + element + '>,' +\
        'std::vector<' + element + '>,' +\
        'std::vector<' + element + '>::iterator' +\
      '>(' + str(n) + ', ' + str(k) + ')'
    self.time_unit = 'ns'
    self.driver_file = self.generate_cpu_time_driver()

  def output(self):
    return (self.n, float(self.driver_output) / self.n)

class heapsort_case(benchmark.case):
  def __init__(self, n, k, element, experiment, d, policy, heap_functions):
    benchmark.case.__init__(self)
    self.n = n
#   self.computer = 'pc-014.diku.dk'
#   self.compiler = '/scratch/g++3.0.4/bin/g++'
#    self.compiler = '/usr/bin/g++-3.2'     
    self.compiler = 'g++-3' 
    self.compiler_options.extend(['-O6'])
    home = os.environ['HOME']
    self.include_paths.extend([
      home + '/CPHSTL/Report/In-place-multiway-heaps/Program/Benchmark/',
      home + '/CPHSTL/Report/In-place-multiway-heaps/Program/Common/',
      home + '/CPHSTL/Report/In-place-multiway-heaps/Program/' + policy + '/',
      home + '/CPHSTL/Report/In-place-multiway-heaps/Program/' + heap_functions + '/' 
      ])
    self.include_files.extend(['<vector>', '<functional>', 'heap_policy.cpp', 
      'heap_functions.cpp', experiment + '.c++'])
    self.dual_exists = 1
    self.constructor_call = \
      'experiment<' + \
        str(d) + ', ' + \
         element + ',' + \
         'std::less<' + element + '>,' +\
        'std::vector<' + element + '>,' +\
        'std::vector<' + element + '>::iterator, ' +\
        'heap_policy<' +\
          str(d) + ', ' +\
          'std::vector<' + element + '>::iterator, ' +\
          'std::less<' + element + '> ' +\
        '> ' +\
      '>(' + str(n) + ', ' + str(k) + ')'
    self.time_unit = 'ns'
    self.driver_file = self.generate_cpu_time_driver()

  def output(self):
    return (self.n, float(self.driver_output) / self.n)

#  def tidy_up(self): 
#    "Used for debugging to see the files generated"
#    pass 
  
class cphstlsort_curve(benchmark.curve_suite):
  def __init__(self, element, experiment, heap, heap_functions, log_i, log_k, k):
    benchmark.curve_suite.__init__(self)
    self.title = experiment + ' ' + heap_functions
    for j in range(log_i, log_k + 1):
      self.add(cphstlsort_case(1 << j, k, element, experiment, heap, heap_functions))

class sandersort_curve(benchmark.curve_suite):
  def __init__(self, element, experiment, heap_functions, log_i, log_k, k):
    benchmark.curve_suite.__init__(self)
    self.title = experiment + ' ' + heap_functions
    for j in range(log_i, log_k + 1):
      self.add(sandersort_case(1 << j, k, element, experiment, heap_functions))

class stdsort_curve(benchmark.curve_suite):
  def __init__(self, element, experiment, heap_functions, log_i, log_k, k):
    benchmark.curve_suite.__init__(self)
    self.title = experiment + ' ' + heap_functions
    for j in range(log_i, log_k + 1):
      self.add(stdsort_case(1 << j, k, element, experiment, heap_functions))

class heapsort_curve(benchmark.curve_suite):
  def __init__(self, element, experiment, d, policy, heap_functions, log_i, log_k, k):
    benchmark.curve_suite.__init__(self)
    self.title =  experiment + ' ' + heap_functions + ' ' + policy
    for j in range(log_i, log_k + 1):
      self.add(heapsort_case(1 << j, k, element, experiment, d, policy, heap_functions))
