/*
Benchmark Sanders binary heap function
*/

#include "do_nothing.c++" // defines do_nothing()
#include <climits> // defines INT_MAX
#include "heap2.h"

template <
  typename element, 
  typename ordering,
  typename container,
  typename position
>
class experiment {
public:
  experiment(unsigned int n, unsigned int k) : n(n), k(k) {
    h = container(k, element());
  }

#define rand32() (ran32State = 1664525 * ran32State + 1013904223)
#define getRandom() ((rand32()), (int)(ran32State >> 2))

#define HTYPE Heap2<int, int> 
#define HINIT heap(INT_MAX, -INT_MAX, n)

  inline void onePass(HTYPE& heap, int n, int i)  { 
    int j, newElem, k, v, tmpkey;
    static int ran32State = 42 << 20;
    
    for (j = 0;  j < n;  j++) {   
      newElem = getRandom();
      heap.insert(newElem, j);
    }
    
    for (j = 0;  j < i;  j++) {   
      tmpkey = heap.getMinKey();
      heap.deleteMin(&k, &v);
      
      newElem = tmpkey + h[i];
      heap.insert(newElem, j);
    }
  }

  void primal() {
    HTYPE HINIT;
    random_sequence_for_hold(h.begin(), h.end());
    onePass(heap, n, k);
  }
  
  void dual() {
    HTYPE HINIT;
    random_sequence_for_hold(h.begin(), h.end());
    do_nothing<void, HTYPE&, unsigned int>(heap, n);
  }
private:
  unsigned int n;
  unsigned int k;
  container h; // hold data
};
