#include <algorithm> // defines std::make_heap
#include <vector> // define std::vector
#include <iterator> // defines std::iterator_traits
#include <functional> // defines std::less
#include "input_generation.c++" // defines random_sequence()
#include "do_nothing.c++" // defines do_nothing()
#include <cstdlib> // defines std::rand and std::srand

template <
  typename element, 
  typename ordering,
  typename container,
  typename position
>
class experiment {
public:
  experiment(unsigned int n) : n(n) {
    a = container(n, element());
    less = ordering();
  }

  template <position, ordering>
  void 
  onePass(position first, position past_the_end, ordering less)  { 
    typedef typename std::iterator_traits<position>::value_type value_type;
    position q;

    for (q = first; q != past_the_end; ++q) {
      *q = value_type(unsigned(std::rand())); 
      std::push_heap(first, first + q, less);

      std::pop_heap(first, first + q, less);

      *q = value_type(unsigned(std::rand())); 
      std::push_heap(first, first + q, less);
    }
    
    for (q = first; q != past_the_end; ++q) {
      std::pop_heap(first, first + q, less);
      
      *q = value_type(unsigned(std::rand())); 
      std::push_heap(first, first + q, less);

      std::pop_heap(first, first + q, less);
    }
  }

  void primal() {
    onePass<position, ordering>(a.begin(), a.end(), less);
  }
  void dual() {
    do_nothing<void, position, position, ordering>(a.begin(), a.end(), less);
  }
private:
  unsigned int n;
  container a;
  ordering less;
};

