"""
Benchmarking various versions of heapsort

shell> echo $PYTHONPATH
/home/disk04/jyrki/CPHSTL/Tool/Benchmark/
"""

__author__ = "Jyrki Katajainen"
__email__ = "jyrki@diku.dk"
__version__ = "$Revision: 1.6 $"[11:-2]
__date__ = "$Date: 2004/01/06 16:40:23 $"[7:-2]

import benchmark
import os

class stdsort_case(benchmark.case):
  def __init__(self, n, element, function, directory):
    benchmark.case.__init__(self)
    self.n = n
#    self.computer = 'pc-014.diku.dk'
#    self.compiler = '/scratch/g++3.0.4/bin/g++'
    self.compiler = '/usr/bin/g++-3.2.2' 
#    self.compiler = '/usr/bin/g++-3.2' 
    self.compiler_options.extend(['-O6'])
    home = os.environ['HOME']
    self.include_paths.extend([
      home + '/CPHSTL/Report/In-place-multiway-heaps/Program/Common/',
      home + '/CPHSTL/Report/In-place-multiway-heaps/Program/' + directory + '/' 
      ])
    self.include_files.extend(['<vector>', '<functional>', function + '.c++'])
    self.dual_exists = 1
    self.constructor_call = \
      'experiment<' + \
         element + ',' + \
         'std::less<' + element + '>,' +\
        'std::vector<' + element + '>,' +\
        'std::vector<' + element + '>::iterator' +\
      '>(' + str(n) + ')'
    self.time_unit = 'ns'
    self.driver_file = self.generate_cpu_time_driver()

  def output(self):
    return (self.n, float(self.driver_output) / self.n)


class briansort_case(benchmark.case):
  def __init__(self, n, element, function, directory):
    benchmark.case.__init__(self)
    self.n = n
#    self.computer = 'pc-014.diku.dk'
#    self.compiler = '/scratch/g++3.0.4/bin/g++'
    self.compiler = '/usr/bin/g++-3.2.2' 
#    self.compiler = '/usr/bin/g++-3.2' 
    self.compiler_options.extend(['-O6'])
    home = os.environ['HOME']
    self.include_paths.extend([
      home + '/CPHSTL/Report/In-place-multiway-heaps/Program/Common/',
      home + '/CPHSTL/Report/In-place-multiway-heaps/Program/' + directory + '/' 
      ])
    self.include_files.extend(['cphstl_8heap.cpp', '<vector>', '<functional>', function + '.c++'])
    self.dual_exists = 1
    self.constructor_call = \
      'experiment<' + \
         element + ',' + \
         'std::less<' + element + '>,' +\
        'std::vector<' + element + '>,' +\
        'std::vector<' + element + '>::iterator' +\
      '>(' + str(n) + ')'
    self.time_unit = 'ns'
    self.driver_file = self.generate_cpu_time_driver()

  def output(self):
    return (self.n, float(self.driver_output) / self.n)

class sandersort_case(benchmark.case):
  def __init__(self, n, element, function, directory):
    benchmark.case.__init__(self)
    self.n = n
#    self.computer = 'pc-014.diku.dk'
#    self.compiler = '/scratch/g++3.0.4/bin/g++'
    self.compiler = '/usr/bin/g++-3.2.2' 
#    self.compiler = '/usr/bin/g++-3.2' 
    self.compiler_options.extend(['-O6'])
    home = os.environ['HOME']
    self.include_paths.extend([
      home + '/CPHSTL/Report/In-place-multiway-heaps/Program/Common/',
      home + '/CPHSTL/Report/In-place-multiway-heaps/Program/' + directory + '/' 
      ])
    self.include_files.extend(['<vector>', '<functional>', function + '.c++'])
    self.dual_exists = 1
    self.constructor_call = \
      'experiment<' + \
         element + ',' + \
         'std::less<' + element + '>,' +\
        'std::vector<' + element + '>,' +\
        'std::vector<' + element + '>::iterator' +\
      '>(' + str(n) + ')'
    self.time_unit = 'ns'
    self.driver_file = self.generate_cpu_time_driver()

  def output(self):
    return (self.n, float(self.driver_output) / self.n)

class heapsort_case(benchmark.case):
  def __init__(self, n, element, experiment, d, policy, heap_functions):
    benchmark.case.__init__(self)
    self.n = n
#   self.computer = 'pc-014.diku.dk'
#   self.compiler = '/scratch/g++3.0.4/bin/g++'
    self.compiler = '/usr/bin/g++-3.2.2' 
#    self.compiler = '/usr/bin/g++-3.2' 
    self.compiler_options.extend(['-O6'])
    home = os.environ['HOME']
    self.include_paths.extend([
      home + '/CPHSTL/Report/In-place-multiway-heaps/Program/Benchmark/',
      home + '/CPHSTL/Report/In-place-multiway-heaps/Program/Common/',
      home + '/CPHSTL/Report/In-place-multiway-heaps/Program/' + policy + '/',
      home + '/CPHSTL/Report/In-place-multiway-heaps/Program/' + heap_functions + '/' 
      ])
    self.include_files.extend(['<vector>', '<functional>', 'heap_policy.cpp', 
      'heap_functions.cpp', experiment + '.c++'])
    self.dual_exists = 1
    self.constructor_call = \
      'experiment<' + \
        str(d) + ', ' + \
         element + ',' + \
         'std::less<' + element + '>,' +\
        'std::vector<' + element + '>,' +\
        'std::vector<' + element + '>::iterator, ' +\
        'heap_policy<' +\
          str(d) + ', ' +\
          'std::vector<' + element + '>::iterator, ' +\
          'std::less<' + element + '> ' +\
        '> ' +\
      '>(' + str(n) + ')'
    self.time_unit = 'ns'
    self.driver_file = self.generate_cpu_time_driver()

  def output(self):
    return (self.n, float(self.driver_output) / self.n)

#  def tidy_up(self): 
#    "Used for debugging to see the files generated"
#    pass 
  
class briansort_curve(benchmark.curve_suite):
  def __init__(self, element, experiment, heap_functions, log_i, log_k):
    benchmark.curve_suite.__init__(self)
    self.title = 'experiment: ' + experiment + ' source: ' + heap_functions 
    for j in range(log_i, log_k + 1):
      self.add(briansort_case(1 << j, element, experiment, heap_functions))

class sandersort_curve(benchmark.curve_suite):
  def __init__(self, element, experiment, function, log_i, log_k):
    benchmark.curve_suite.__init__(self)
    self.title = 'experiment: ' + experiment + ' source: ' + function
    for j in range(log_i, log_k + 1):
      self.add(sandersort_case(1 << j, element, experiment, function))

class stdsort_curve(benchmark.curve_suite):
  def __init__(self, element, experiment, heap_functions, log_i, log_k):
    benchmark.curve_suite.__init__(self)
    self.title = 'experiment: ' + experiment + ' source: ' + heap_functions 
    for j in range(log_i, log_k + 1):
      self.add(stdsort_case(1 << j, element, experiment, heap_functions))

class heapsort_curve(benchmark.curve_suite):
  def __init__(self, element, experiment, d, policy, heap_functions, log_i, log_k):
    benchmark.curve_suite.__init__(self)
    self.title = 'experiment: ' + experiment + ' source: ' + heap_functions + \
                 ' policy: ' + policy + ' arity: ' + str(d)
    for j in range(log_i, log_k + 1):
      self.add(heapsort_case(1 << j, element, experiment, d, policy, heap_functions))

class sort_plot(benchmark.plot_suite):
  def __init__(self, element, log_i, log_k):
    benchmark.plot_suite.__init__(self)
    self.title = 'Efficiency of Sanders heap functions' 
    self.xlabel = 'n'
    self.ylabel = 'Execution time per element [in nanoseconds]'
#    self.add(heapsort_curve(element, 'heapsort', 2, 'In-place-policy', 'Top-down', log_i, log_k))
#    self.add(heapsort_curve(element, 'heapsort', 4, 'In-place-policy', 'Top-down', log_i, log_k))
#    self.add(heapsort_curve(element, 'heapsort', 8, 'In-place-policy', 'Top-down', log_i, log_k))
#    self.add(heapsort_curve(element, 'heapsort', 16, 'In-place-policy', 'Top-down', log_i, log_k))
#    self.add(heapsort_curve(element, 'heapsort', 2, 'In-place-2-policy', 'Top-down', log_i, log_k))
#    self.add(heapsort_curve(element, 'heapsort', 2, 'In-place-2-policy', 'Bottom-up', log_i, log_k))
#    self.add(heapsort_curve(element, 'heapsort', 2, 'In-place-2-policy', 'Bottom-up_move-saving', log_i, log_k))
#    self.add(heapsort_curve(element, 'heapsort', 4, 'In-place-4-policy', 'Top-down', log_i, log_k))
#    self.add(heapsort_curve(element, 'heapsort', 4, 'In-place-4-policy', 'Bottom-up', log_i, log_k))
#    self.add(heapsort_curve(element, 'heapsort', 4, 'In-place-4-policy', 'Bottom-up_move-saving', log_i, log_k))
#    self.add(heapsort_curve(element, 'heapsort', 4, 'In-place-4-policy-nested-If', 'Top-down', log_i, log_k))
#    self.add(heapsort_curve(element, 'heapsort', 4, 'In-place-4-policy-unnested-If', 'Top-down', log_i, log_k))
#    self.add(heapsort_curve(element, 'heapsort', 4, 'In-place-4-policy-unnested-Sp', 'Top-down', log_i, log_k))
#    self.add(heapsort_curve(element, 'heapsort', 4, 'In-place-4-policy-unnested-Sp-if', 'Top-down', log_i, log_k))
#    self.add(heapsort_curve(element, 'heapsort', 4, 'In-place-4-policy-unnested-Sp-Seriel', 'Top-down', log_i, log_k))
#    self.add(heapsort_curve(element, 'heapsort', 4, 'In-place-4-policy-unnested-If-Sander', 'Top-down', log_i, log_k))
#    self.add(heapsort_curve(element, 'heapsort', 4, 'In-place-4-policy-unnested-If-Sander-style', 'Top-down', log_i, log_k))
#    self.add(heapsort_curve(element, 'heapsort', 8, 'In-place-8-policy', 'Top-down', log_i, log_k))
#    self.add(heapsort_curve(element, 'heapsort', 8, 'In-place-8-policy', 'Bottom-up', log_i, log_k))
#    self.add(heapsort_curve(element, 'heapsort', 8, 'In-place-8-policy', 'Bottom-up_move-saving', log_i, log_k))
#    self.add(heapsort_curve(element, 'heapsort', 8, 'In-place-8-policy-unnested-If', 'Top-down', log_i, log_k))
#    self.add(heapsort_curve(element, 'heapsort', 8, 'In-place-8-policy-unnested-Sp', 'Top-down', log_i, log_k))
#    self.add(heapsort_curve(element, 'heapsort', 8, 'In-place-8-policy-unnested-If-Seriel', 'Top-down', log_i, log_k))
#    self.add(heapsort_curve(element, 'heapsort', 16, 'In-place-16-policy', 'Top-down', log_i, log_k))
#    self.add(heapsort_curve(element, 'heapsort', 16, 'In-place-16-policy', 'Bottom-up', log_i, log_k))
#    self.add(heapsort_curve(element, 'heapsort', 16, 'In-place-16-policy', 'Bottom-up_move-saving', log_i, log_k))
#    self.add(briansort_curve(element, 'sort_heap', 'Brian4', log_i, log_k))
#    self.add(briansort_curve(element, 'sort_heap', 'Brian', log_i, log_k))
#    self.add(briansort_curve(element, 'sort_heap', 'Brian16', log_i, log_k))
#    self.add(briansort_curve(element, 'sort_heap', 'Brian-Sp-Seriel', log_i, log_k))
#    self.add(briansort_curve(element, 'sort_heap', 'Brian-Sp-Seriel-While', log_i, log_k))
#    self.add(briansort_curve(element, 'sort_heap', 'Brian-SpSp-Seriel', log_i, log_k))
    self.add(sandersort_curve(element, 'sander_sort_heap2', 'Sander', log_i, log_k))
    self.add(sandersort_curve(element, 'sander_sort_heap4', 'Sander', log_i, log_k))
#    self.add(sandersort_curve(element, 'sander_insert_delete_heap2', 'Sander', log_i, log_k))
#    self.add(sandersort_curve(element, 'sander_insert_delete_heap4', 'Sander', log_i, log_k))
#    self.add(sandersort_curve(element, 'sander_hold_heap2', 'Sander', log_i, log_k))
#    self.add(sandersort_curve(element, 'sander_hold_heap4', 'Sander', log_i, log_k))
    self.add(stdsort_curve(element, 'sort', 'SGI', log_i, log_k))
    self.add(stdsort_curve(element, 'partial_sort', 'SGI', log_i, log_k))
    self.gnuplot_commands += """
set title '%(title)s'
set xlabel '%(xlabel)s'
set ylabel '%(ylabel)s'
set logscale x
""" % self.__dict__

if __name__ == '__main__':
  benchmark.main(
    task = sort_plot('int', 10, 23),
    runner = benchmark.gnuplot_runner
  )
