#ifndef __UTILITY_FUNCTIONS__
#define __UTILITY_FUNCTIONS__

#include <iostream>

template <int d, typename position, typename level, typename index, typename element, 
  typename ordering, typename policy>
index
correct_place(position a, index l, level h, element x, const ordering& less, policy& p) {
  
  /* bin. part: i, j, and h needed */

  index i = p.root();
  index j = l;        
  index k = j;

  while (h > 1) {
    level g = h / 2;
    j = p.ancestor(k, h, g);
    if (less(a[j], x)) {
      h = h - g;
      k = j;
    }
    else {
      h = g;
      i = j;
    }
  }
  
  if (less(x, a[i])) {
    return k;
  }
  else {
    return i;
  }
}

template <int d, typename position, typename index, typename element, 
  typename ordering, typename policy>
void
sift_down(position a, index j, element x, const ordering& less, policy& p) {
  //std::cout << "before sift_down j,x: " << j << " " << x << std::endl;
  //show(a, a + p.size());
  
  typename policy::level h = p.depth(p.last_leaf()) - p.depth(j);
  while (h > 1) {
  
    index i = j;  
    typename policy::level g = 1;
    while (g <= h / 2) {
      index k = p.top_all_present(a, j, less);
      p.update(a, j, a[k]);
      j = k;
      g = g + 1;
    }
    if (less(x, a[j])) {
      h = h - h / 2;
      i = j;
    }
    else {
      while (j != i && less(a[p.parent(j)], x)) {
	p.update(a, j, a[p.parent(j)]);
	j = p.parent(j);
      }
      p.update(a, j, x);
      return;
    }
  }
  index k = p.first_child(j);
  if (k < p.size()) {
    if (k + d -1 < p.size()) {
      k = p.top_all_present(a, j, less);
    }
    else {
      k = p.top_some_absent(a, j, less);
    }
    if (less(x, a[k])) {
      p.update(a, j, a[k]);
      p.update(a, k, x);
    }
    else {
      p.update(a, j, x);
    }
  }
  else {
    p.update(a, j, x);
  }

  //std::cout << "after sift_down: " << std::endl;
  //show(a, a + p.size());
} 

template <int d, typename position, typename index, typename element, 
  typename ordering, typename policy>
void
sift_up(position a, index j, element x, const ordering& less, policy& p) {
  //std::cout << "before sift_up j: " << j << std::endl;
  //show(a, a + p.size());
  
  typedef typename policy::level level; 
  level h;
  index i;

  h = p.depth(j);

  i = correct_place<d, position, level, index, element, ordering, policy>(a, j, h, x, less, p);

  while (i != j) {
    p.update(a, j, a[p.parent(j)]);
    j = p.parent(j);
  }
  
  p.update(a, i, x);
  
  //std::cout << "after sift_up" << std::endl;
  //show(a, a + p.size());
} 

#endif // __UTILITY_FUNCTIONS__


