#ifndef __UTILITY_FUNCTIONS__
#define __UTILITY_FUNCTIONS__

#include <iostream>

template <int d, typename position, typename level, typename index, typename element,
	  typename ordering, typename policy>
index
correct_place(position a, index l, level h, element x, const ordering& less, policy& p) {

  /* exp. part: parameters needed */

  if (h == 0) return l;

  level delta = 1;
  index i;
  index j = l;
  while (true) {
    if (delta > h) {
      i = p.ancestor(j, h, h);
      break;
    }
    i = p.ancestor(j, h, delta);
    if (less(a[i], x)) {
      h = h - delta;
      delta = 2 * (delta + 1) - 1;
      j = i;
    }
    else {
      h = delta;
      break;
    }
  }

  /* bin. part: i, j, and h needed */

  index k = j;
  while (h > 1) {
    level g = h / 2;
    j = p.ancestor(k, h, g);
    if (less(a[j], x)) {
      h = h - g;
      k = j;
    }
    else {
      h = g;
      i = j;
    }
  }

  if (less(x, a[i])) {
    return k;
  }
  else {
    return i;
  }
}

template <int d, typename position, typename index, typename element,
  typename ordering, typename policy>
void
sift_down(position a, index i, element x, const ordering& less, policy& p) {
  //std::cout << "before sift_down i,x: " << i << " " << x << std::endl;
  //show(a, a + p.size());
  index m = i;
  index j = p.first_child(i);
  const index last_full_branch = p.parent(p.last_leaf() + 1) - 1;

  while (i <= last_full_branch) {
    j = p.top_all_present(a, i, less);
    p.update(a, i, a[j]);
    i = j;
  }

  j = p.first_child(i);
  if (j < p.size()) {
    j = p.top_some_absent(a, i, less);
    p.update(a, i, a[j]);
    if (less(a[i], x)) {
      p.update(a, j, a[i]);
    }
    else {
      p.update(a, j, x);
      return;
    }
  }

  while ((i != m) && less(a[p.parent(i)], x)) {
    p.update(a, i, a[p.parent(i)]);
    i = p.parent(i);
  }
  p.update(a, i, x);

 //std::cout << "after sift_down: " << std::endl;
 //show(a, a + p.size());
}


#include <vector> // defines std::vector

template <int d, typename position, typename index, typename element,
  typename ordering, typename policy>
void
sift_up_buffer(position a, const ordering& less, policy& p) {
  //std::cout << "before sift_up_buffer" << std::endl;
  //show(a, a + p.size());

  typedef typename policy::level level;
  index i;
  index l = p.last_leaf() + 1;
  index buffer_parent = p.parent(l);
  level hl = p.depth(l);
  element tmp_elem;
  index j = p.top_all_present(a, buffer_parent, less);   
  index c;
  index hr;
  hr = p.depth(p.root());
  c = correct_place<d, position, level, index, element, ordering, policy>(a, j, hr, a[j], less, p);
  level hc = p.depth(c);

  do {
    i = p.ancestor(l, hl, hc);
    if (less(a[i], a[j])) {
      tmp_elem = a[i];
      a[i] = a[j];
      p.update(a, j, tmp_elem);
      j = p.top_all_present(a, buffer_parent, less);   
    }
    hc--;
  } while (hc >= 0);

  //std::cout << "after sift_up_buffer" << std::endl;
  //show(a, a + p.size());
 }

#endif // __UTILITY_FUNCTIONS__
