#ifndef __UTILITY_FUNCTIONS__
#define __UTILITY_FUNCTIONS__

#include <iostream>

template <int d, typename position, typename index, typename ordering, 
  typename policy>
typename policy::index
traverse_down(position a, index i, const ordering& less, const policy& p) {
  const index last_full_branch = p.parent(p.last_leaf() + 1) - 1;
  
  while (i <= last_full_branch) {
    i = p.top_all_present(a, i, less);
  }
  if (p.first_child(i) <= p.last_leaf()) {
    i = p.top_some_absent(a, i, less);
  }
  return i;
}

template <int d, typename position, typename index, typename element, 
  typename ordering, typename policy>
typename policy::index
bottom_up_search(position a, index i, index k, const element& x, const ordering& less,
  const policy& p) {
  while (k != i && !(less(x, *(a + k)))) {
    k = p.parent(k);
  }
  return k;
}

template <int d, typename position, typename index, typename element, 
  typename ordering, typename policy>
void
lift_up(position a, index i, index k, const element& x, policy& p) {
  typename policy::level h = p.depth(k);  
  typename policy::level Delta = h - p.depth(i);
 
  while (Delta > 0) {
    --Delta;
    typename policy::index j = p.ancestor(k, h, Delta);
    p.update(a, i, *(a + j));
    i = j;
  }
  p.update(a, k, x);
}

template <typename position, typename index, typename element, typename ordering,
  typename policy>
typename policy::index
top_down_search(position a, index i, index k, const element& x, const ordering& less,
  const policy& p) {
  typename policy::level h = p.depth(k);  
  typename policy::level Delta = h - p.depth(i);
  while (Delta > 0 && !(less(a[i], x))) {
    --Delta;
    i = p.ancestor(k, h, Delta);
  }
  return i;
}

template <int d, typename position, typename index, typename element, 
  typename ordering, typename policy>
void
lift_down(position a, index i, index k, const element& x, policy& p) {
  typename policy::index j;

  while (i != k) {
    j = p.parent(k);
    p.update(a, k, a[j]);
    k = j;
  }
  p.update(a, k, x);
}

template <int d, typename position, typename index, typename element, 
  typename ordering, typename policy>
void
sift_down(position a, index i, element x, const ordering& less, policy& p) {
  //std::cout << "before traverse i: " << i << std::endl;
  //show(a, a + p.size());
  index k = traverse_down<d, position, index, ordering, policy>(a, i, less, p);
  //std::cout << "before search i,k: " << i << " " << k << std::endl;
  //show(a, a + p.size());
  index j = bottom_up_search<d, position, index, element, ordering, policy>(a, i, k, x, less, p);
  //std::cout << "before lift_up j: " << j << std::endl;
  //show(a, a + p.size());
  lift_up<d, position, index, element, ordering, policy>(a, i, j, x, p);
  //std::cout << "after lift_up j: " << j << std::endl;
  //show(a, a + p.size());
} 

template <int d, typename position, typename index, typename element, 
  typename ordering, typename policy>
void
sift_up(position a, index j, element x, const ordering& less, policy& p) {
  //std::cout << "before search j: " << j << std::endl;
  //show(a, a + p.size());
  typename policy::index i = top_down_search(a, p.root(), j, x, less, p);
  //std::cout << "before lift_down i: " << i << std::endl;
  //show(a, a + p.size());
  lift_down<d, position, index, element, ordering, policy>(a, i, j, x, p);
  //std::cout << "after lift_down" << std::endl;
  //show(a, a + p.size());
} 

#endif // __UTILITY_FUNCTIONS__
