#ifndef __HEAP_POLICY__
#define __HEAP_POLICY__

#include "heap_policy.hpp" // declares heap_policy
#include <cmath> // defines std::ilogb

extern int ilogb(double) throw();

#ifdef BUFFER
template <int d, typename position, typename ordering>
heap_policy<d, position, ordering>::heap_policy (
  heap_policy<d, position, ordering>::index n
): n(n) {
  if (n <= index(d) + 1) { l = 1;}
  else {
    l = d * ((n - 1) / index(d) - 1) + 1;
    if ( l + index(d) != n) {
      l = l + index(d);
    }
  }
}
#else
template <int d, typename position, typename ordering>
heap_policy<d, position, ordering>::heap_policy (
  heap_policy<d, position, ordering>::index n
): n(n) {
}
#endif

template <int d, typename position, typename ordering>
bool
heap_policy<d, position, ordering>::is_root (
  heap_policy<d, position, ordering>::index i
) const {
  return (i == 0);
}

template <int d, typename position, typename ordering>
bool
heap_policy<d, position, ordering>::is_first_child (
  heap_policy<d, position, ordering>::index i
) const {
  typedef typename heap_policy<d, position, ordering>::index index;
  return (i - (i / index(d)) * index(d) == 1);
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::size () const {
  return n;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::level
heap_policy<d, position, ordering>::depth (
  heap_policy<d, position, ordering>::index i
) const {
  typedef typename heap_policy<d, position, ordering>::level level;
  typedef typename heap_policy<d, position, ordering>::index index;
  level h = 0;
  index m = 1;
  index nr = 0;
  while (i > nr) {
    m = m * index(d);
    nr = nr + m;
    ++h;
  }
  return h;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::root () const {
  return 0;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::leftmost_leaf () const {
  typedef typename heap_policy<d, position, ordering>::level level;
  typedef typename heap_policy<d, position, ordering>::index index;
  index m = heap_policy<d, position, ordering>::size() - 1;
  level h = heap_policy<d, position, ordering>::depth(m);
  index r = 1;
  while (h > 0) {
    r = r * index(d);
    --h;
  }
  r = (r - 1) / index(d - 1);
  return r;
}

#ifdef BUFFER
template <int d, typename position, typename ordering>
heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::last_leaf () const {
  return l - 1;
}
#else
template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::last_leaf () const {
  return n - 1;
}
#endif

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::first_child (
  heap_policy<d, position, ordering>::index p
) const {
  return d * p + 1;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::parent (
  heap_policy<d, position, ordering>::index c
) const {
  return (c - 1) / d;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::ancestor (
  heap_policy<d, position, ordering>::index j,
  heap_policy<d, position, ordering>::level Delta
) const {
  typedef typename heap_policy<d, position, ordering>::index index;
  index t = 1;
  while (Delta > 0) {
    t = t * index(d);
    --Delta;
  }
  index i = (t - 1) / index(d - 1);
  return (j - i) / t;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::top_some_absent (
  position a,
  heap_policy<d, position, ordering>::index i,
  const ordering& less
) const {
  typedef typename heap_policy<d, position, ordering>::index index;
  typedef typename heap_policy<d, position, ordering>::element element;
  index j = heap_policy<d, position, ordering>::first_child(i);
  index k = (j + index(d) < n) ? j + index(d) : n;
  index maxc = j;
  element max = a[j];
  ++j;
  while (j < k) {
    if (less(max, a[j])) {
      maxc = j;
      max = a[j];
    }
    ++j;
  }
  return maxc;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::top_all_present (
  position a,
  heap_policy<d, position, ordering>::index i,
  const ordering& less
) const {
  return heap_policy<d, position, ordering>::top_some_absent(a, i, less);
}

template <int d, typename position, typename ordering>
void
heap_policy<d, position, ordering>::update (
  position a,
  heap_policy<d, position, ordering>::index i,
  const heap_policy<d, position, ordering>::element& x
) {
  a[i] = x;
}

#ifdef BUFFER
template <int d, typename position, typename ordering>
void
heap_policy<d, position, ordering>::erase_last_leaf (
) {
  --n;
  if (l == n) {
    if ( n <= index(d) ) {l = 1;}
    else {l = l - index(d);}
  }
}
#else
template <int d, typename position, typename ordering>
void
heap_policy<d, position, ordering>::erase_last_leaf (
) {
  --n;
}
#endif

#ifdef BUFFER
template <int d, typename position, typename ordering>
void
heap_policy<d, position, ordering>::insert_new_leaf (
) {
  ++n;
  if (( l + 2 * index(d)) == n ) {
    l = l + index(d);
  }
}
#else
template <int d, typename position, typename ordering>
void
heap_policy<d, position, ordering>::insert_new_leaf (
) {
  ++n;
}
#endif


/*

Specialization for d = 16.

*/

#ifdef BUFFER
template <> template <typename position, typename ordering>
heap_policy<16, position, ordering>::heap_policy (
  heap_policy<16, position, ordering>::index n
): n(n) {
  if (n <= 17) { l = 1;}
  else {
    l = 16 * ((n - 1) / 16 - 1) + 1;
    if ( l + 16 != n) {
      l = l + 16;
    }
  }
}
#else
template <> template <typename position, typename ordering>
heap_policy<16, position, ordering>::heap_policy (
  heap_policy<16, position, ordering>::index n
): n(n) {
}
#endif

template <> template <typename position, typename ordering>
bool
heap_policy<16, position, ordering>::is_root (
  heap_policy<16, position, ordering>::index i
) const {

  return (i == 0);
}

template <> template <typename position, typename ordering>
bool
heap_policy<16, position, ordering>::is_first_child (
  heap_policy<16, position, ordering>::index i
) const {

  return (i & 15) == 1;
}

template <> template <typename position, typename ordering>
typename heap_policy<16, position, ordering>::index
heap_policy<16, position, ordering>::size () const {

  return n;
}

template <> template <typename position, typename ordering>
typename heap_policy<16, position, ordering>::level
heap_policy<16, position, ordering>::depth (
  heap_policy<16, position, ordering>::index i
  ) const {
  typedef typename heap_policy<16, position, ordering>::level level;

  if (i == 0) return 0;
  if (i < 9) return 1;
  return level(ilogb(15 * i + 1) >> 2);
}

template <> template <typename position, typename ordering>
typename heap_policy<16, position, ordering>::index
heap_policy<16, position, ordering>::root () const {

  return 0;
}

template <> template <typename position, typename ordering>
typename heap_policy<16, position, ordering>::index
heap_policy<16, position, ordering>::leftmost_leaf () const {
  typedef unsigned int element;
  index ll = heap_policy<16, position, ordering>::size() - 1;
  const static element lml[12] = {
    0, 1, 17, 273, 4369, 69905, 1118481, 17895697, 286331153
  };
 return lml[depth(ll)];
}

#ifdef BUFFER
template <> template <typename position, typename ordering>
typename heap_policy<16, position, ordering>::index
heap_policy<16, position, ordering>::last_leaf () const {
  return l - 1;
}
#else
template <> template <typename position, typename ordering>
typename heap_policy<16, position, ordering>::index
heap_policy<16, position, ordering>::last_leaf () const {
  return n - 1;
}
#endif

template <> template <typename position, typename ordering>
typename heap_policy<16, position, ordering>::index
heap_policy<16, position, ordering>::first_child (
  heap_policy<16, position, ordering>::index p
) const {

  return (p << 4) + 1;
}

template <> template <typename position, typename ordering>
typename heap_policy<16, position, ordering>::index
heap_policy<16, position, ordering>::parent (
  heap_policy<16, position, ordering>::index c
) const {

  return (c - 1) >> 4;
}

template <> template <typename position, typename ordering>
typename heap_policy<16, position, ordering>::index
heap_policy<16, position, ordering>::ancestor (
  heap_policy<16, position, ordering>::index j,
  heap_policy<16, position, ordering>::level Delta
) const {
  typedef typename heap_policy<16, position, ordering>::index index;
  typedef unsigned int element;
  const static element lml[12] = {
    0, 1, 17, 273, 4369, 69905, 1118481, 17895697, 286331153
  };
  index m;

  index d = depth(j);
  if ((d - Delta) < 0) return 0;
  if (Delta == 0) return j;
  m = 16 << (Delta - 1) * 4;
  return ((j - lml[d]) / m) + lml[d - Delta];
}

template <> template <typename position, typename ordering>
typename heap_policy<16, position, ordering>::index
heap_policy<16, position, ordering>::top_some_absent (
  position a,
  heap_policy<16, position, ordering>::index i,
  const ordering& less
) const {
  typedef typename heap_policy<16, position, ordering>::index index;
  index j = heap_policy<16, position, ordering>::first_child(i);
  index ll = heap_policy<16, position, ordering>::size() - 1;
  index r = ll - j;
  index top;

  top = j;

  if (r == 15) {
  if (less(a[top], a[j + 1])) top = j + 1 ;
  if (less(a[top], a[j + 2])) top = j + 2 ;
  if (less(a[top], a[j + 3])) top = j + 3 ;
  if (less(a[top], a[j + 4])) top = j + 4 ;
  if (less(a[top], a[j + 5])) top = j + 5 ;
  if (less(a[top], a[j + 6])) top = j + 6 ;
  if (less(a[top], a[j + 7])) top = j + 7 ;
  if (less(a[top], a[j + 8])) top = j + 8 ;
  if (less(a[top], a[j + 9])) top = j + 9 ;
  if (less(a[top], a[j + 10])) top = j + 10 ;
  if (less(a[top], a[j + 11])) top = j + 11 ;
  if (less(a[top], a[j + 12])) top = j + 12 ;
  if (less(a[top], a[j + 13])) top = j + 13 ;
  if (less(a[top], a[j + 14])) top = j + 14 ;
  if (less(a[top], a[j + 15])) top = j + 15 ;
  return top;
  }

  if (r == 14) {
  if (less(a[top], a[j + 1])) top = j + 1 ;
  if (less(a[top], a[j + 2])) top = j + 2 ;
  if (less(a[top], a[j + 3])) top = j + 3 ;
  if (less(a[top], a[j + 4])) top = j + 4 ;
  if (less(a[top], a[j + 5])) top = j + 5 ;
  if (less(a[top], a[j + 6])) top = j + 6 ;
  if (less(a[top], a[j + 7])) top = j + 7 ;
  if (less(a[top], a[j + 8])) top = j + 8 ;
  if (less(a[top], a[j + 9])) top = j + 9 ;
  if (less(a[top], a[j + 10])) top = j + 10 ;
  if (less(a[top], a[j + 11])) top = j + 11 ;
  if (less(a[top], a[j + 12])) top = j + 12 ;
  if (less(a[top], a[j + 13])) top = j + 13 ;
  if (less(a[top], a[j + 14])) top = j + 14 ;
  return top;
  }

  if (r == 13) {
  if (less(a[top], a[j + 1])) top = j + 1 ;
  if (less(a[top], a[j + 2])) top = j + 2 ;
  if (less(a[top], a[j + 3])) top = j + 3 ;
  if (less(a[top], a[j + 4])) top = j + 4 ;
  if (less(a[top], a[j + 5])) top = j + 5 ;
  if (less(a[top], a[j + 6])) top = j + 6 ;
  if (less(a[top], a[j + 7])) top = j + 7 ;
  if (less(a[top], a[j + 8])) top = j + 8 ;
  if (less(a[top], a[j + 9])) top = j + 9 ;
  if (less(a[top], a[j + 10])) top = j + 10 ;
  if (less(a[top], a[j + 11])) top = j + 11 ;
  if (less(a[top], a[j + 12])) top = j + 12 ;
  if (less(a[top], a[j + 13])) top = j + 13 ;
  return top;
  }

  if (r == 12) {
  if (less(a[top], a[j + 1])) top = j + 1 ;
  if (less(a[top], a[j + 2])) top = j + 2 ;
  if (less(a[top], a[j + 3])) top = j + 3 ;
  if (less(a[top], a[j + 4])) top = j + 4 ;
  if (less(a[top], a[j + 5])) top = j + 5 ;
  if (less(a[top], a[j + 6])) top = j + 6 ;
  if (less(a[top], a[j + 7])) top = j + 7 ;
  if (less(a[top], a[j + 8])) top = j + 8 ;
  if (less(a[top], a[j + 9])) top = j + 9 ;
  if (less(a[top], a[j + 10])) top = j + 10 ;
  if (less(a[top], a[j + 11])) top = j + 11 ;
  if (less(a[top], a[j + 12])) top = j + 12 ;
  return top;
  }

  if (r == 11) {
  if (less(a[top], a[j + 1])) top = j + 1 ;
  if (less(a[top], a[j + 2])) top = j + 2 ;
  if (less(a[top], a[j + 3])) top = j + 3 ;
  if (less(a[top], a[j + 4])) top = j + 4 ;
  if (less(a[top], a[j + 5])) top = j + 5 ;
  if (less(a[top], a[j + 6])) top = j + 6 ;
  if (less(a[top], a[j + 7])) top = j + 7 ;
  if (less(a[top], a[j + 8])) top = j + 8 ;
  if (less(a[top], a[j + 9])) top = j + 9 ;
  if (less(a[top], a[j + 10])) top = j + 10 ;
  if (less(a[top], a[j + 11])) top = j + 11 ;
  return top;
  }

  if (r == 11) {
  if (less(a[top], a[j + 1])) top = j + 1 ;
  if (less(a[top], a[j + 2])) top = j + 2 ;
  if (less(a[top], a[j + 3])) top = j + 3 ;
  if (less(a[top], a[j + 4])) top = j + 4 ;
  if (less(a[top], a[j + 5])) top = j + 5 ;
  if (less(a[top], a[j + 6])) top = j + 6 ;
  if (less(a[top], a[j + 7])) top = j + 7 ;
  if (less(a[top], a[j + 8])) top = j + 8 ;
  if (less(a[top], a[j + 9])) top = j + 9 ;
  if (less(a[top], a[j + 10])) top = j + 10 ;
  if (less(a[top], a[j + 11])) top = j + 11 ;
  return top;
  }

  if (r == 10) {
  if (less(a[top], a[j + 1])) top = j + 1 ;
  if (less(a[top], a[j + 2])) top = j + 2 ;
  if (less(a[top], a[j + 3])) top = j + 3 ;
  if (less(a[top], a[j + 4])) top = j + 4 ;
  if (less(a[top], a[j + 5])) top = j + 5 ;
  if (less(a[top], a[j + 6])) top = j + 6 ;
  if (less(a[top], a[j + 7])) top = j + 7 ;
  if (less(a[top], a[j + 8])) top = j + 8 ;
  if (less(a[top], a[j + 9])) top = j + 9 ;
  if (less(a[top], a[j + 10])) top = j + 10 ;
  return top;
  }

  if (r == 9) {
  if (less(a[top], a[j + 1])) top = j + 1 ;
  if (less(a[top], a[j + 2])) top = j + 2 ;
  if (less(a[top], a[j + 3])) top = j + 3 ;
  if (less(a[top], a[j + 4])) top = j + 4 ;
  if (less(a[top], a[j + 5])) top = j + 5 ;
  if (less(a[top], a[j + 6])) top = j + 6 ;
  if (less(a[top], a[j + 7])) top = j + 7 ;
  if (less(a[top], a[j + 8])) top = j + 8 ;
  if (less(a[top], a[j + 9])) top = j + 9 ;
  return top;
  }

  if (r == 8) {
  if (less(a[top], a[j + 1])) top = j + 1 ;
  if (less(a[top], a[j + 2])) top = j + 2 ;
  if (less(a[top], a[j + 3])) top = j + 3 ;
  if (less(a[top], a[j + 4])) top = j + 4 ;
  if (less(a[top], a[j + 5])) top = j + 5 ;
  if (less(a[top], a[j + 6])) top = j + 6 ;
  if (less(a[top], a[j + 7])) top = j + 7 ;
  if (less(a[top], a[j + 8])) top = j + 8 ;
  return top;
  }

  if (r == 7) {
  if (less(a[top], a[j + 1])) top = j + 1 ;
  if (less(a[top], a[j + 2])) top = j + 2 ;
  if (less(a[top], a[j + 3])) top = j + 3 ;
  if (less(a[top], a[j + 4])) top = j + 4 ;
  if (less(a[top], a[j + 5])) top = j + 5 ;
  if (less(a[top], a[j + 6])) top = j + 6 ;
  if (less(a[top], a[j + 7])) top = j + 7 ;
  return top;
  }

  if (r == 6) {
  if (less(a[top], a[j + 1])) top = j + 1 ;
  if (less(a[top], a[j + 2])) top = j + 2 ;
  if (less(a[top], a[j + 3])) top = j + 3 ;
  if (less(a[top], a[j + 4])) top = j + 4 ;
  if (less(a[top], a[j + 5])) top = j + 5 ;
  if (less(a[top], a[j + 6])) top = j + 6 ;
  return top;
  }

  if (r == 5) {
  if (less(a[top], a[j + 1])) top = j + 1 ;
  if (less(a[top], a[j + 2])) top = j + 2 ;
  if (less(a[top], a[j + 3])) top = j + 3 ;
  if (less(a[top], a[j + 4])) top = j + 4 ;
  if (less(a[top], a[j + 5])) top = j + 5 ;
  return top;
  }

  if (r == 4) {
  if (less(a[top], a[j + 1])) top = j + 1 ;
  if (less(a[top], a[j + 2])) top = j + 2 ;
  if (less(a[top], a[j + 3])) top = j + 3 ;
  if (less(a[top], a[j + 4])) top = j + 4 ;
  return top;
  }

  if (r == 3) {
  if (less(a[top], a[j + 1])) top = j + 1 ;
  if (less(a[top], a[j + 2])) top = j + 2 ;
  if (less(a[top], a[j + 3])) top = j + 3 ;
  return top;
  }

  if (r == 2) {
  if (less(a[top], a[j + 1])) top = j + 1 ;
  if (less(a[top], a[j + 2])) top = j + 2 ;
  return top;
  }

  if (r == 1) {
  if (less(a[top], a[j + 1])) top = j + 1 ;
  return top;
  }

  return top;
}

template <> template <typename position, typename ordering>
typename heap_policy<16, position, ordering>::index
heap_policy<16, position, ordering>::top_all_present (
  position a,
  heap_policy<16, position, ordering>::index i,
  const ordering& less
) const {
  typedef typename heap_policy<16, position, ordering>::index index;
  index j = heap_policy<16, position, ordering>::first_child(i);
  index top;

  top = j;

  less(a[j + 1], a[top]) ? : top = j + 1 ;
  less(a[j + 2], a[top]) ? : top = j + 2 ;
  less(a[j + 3], a[top]) ? : top = j + 3 ;
  less(a[j + 4], a[top]) ? : top = j + 4 ;
  less(a[j + 5], a[top]) ? : top = j + 5 ;
  less(a[j + 6], a[top]) ? : top = j + 6 ;
  less(a[j + 7], a[top]) ? : top = j + 7 ;
  less(a[j + 8], a[top]) ? : top = j + 8 ;
  less(a[j + 9], a[top]) ? : top = j + 9 ;
  less(a[j + 10], a[top]) ? : top = j + 10 ;
  less(a[j + 11], a[top]) ? : top = j + 11 ;
  less(a[j + 12], a[top]) ? : top = j + 12 ;
  less(a[j + 13], a[top]) ? : top = j + 13 ;
  less(a[j + 14], a[top]) ? : top = j + 14 ;
  less(a[j + 15], a[top]) ? : top = j + 15 ;
  return top;

}

template <> template <typename position, typename ordering>
void
heap_policy<16, position, ordering>::update (
  position a,
  heap_policy<16, position, ordering>::index i,
  const heap_policy<16, position, ordering>::element& x
) {
  a[i] = x;
}

#ifdef BUFFER
template <> template <typename position, typename ordering>
void
heap_policy<16, position, ordering>::erase_last_leaf (
) {
  --n;
  if (l == n) {
    if ( n <= 16 ) {l = 1;}
    else {l = l - 16;}
  }
}
#else
template <> template <typename position, typename ordering>
void
heap_policy<16, position, ordering>::erase_last_leaf (
) {
  --n;
}
#endif

#ifdef BUFFER
template <> template <typename position, typename ordering>
void
heap_policy<16, position, ordering>::insert_new_leaf (
) {
  ++n;
  if (( l + 2 * 16) == n ) {
    l = l + 16;
  }
}
#else
template <> template <typename position, typename ordering>
void
heap_policy<16, position, ordering>::insert_new_leaf (
) {

  ++n;
}
#endif

#endif // __HEAP_POLICY__
