#ifndef __HEAP_POLICY__
#define __HEAP_POLICY__

#include "heap_policy.hpp" // declares heap_policy
#include <cmath> // defines std::ilogb

extern int ilogb(double) throw();

#ifdef BUFFER
template <int d, typename position, typename ordering>
heap_policy<d, position, ordering>::heap_policy (
  heap_policy<d, position, ordering>::index n
): n(n) {
  if (n <= index(d) + 1) { l = 1;}
  else {
    l = d * ((n - 1) / index(d) - 1) + 1;
    if ( l + index(d) != n) {
      l = l + index(d);
    }
  }
}
#else
template <int d, typename position, typename ordering>
heap_policy<d, position, ordering>::heap_policy (
  heap_policy<d, position, ordering>::index n
): n(n) {
}
#endif

template <int d, typename position, typename ordering>
bool
heap_policy<d, position, ordering>::is_root (
  heap_policy<d, position, ordering>::index i
) const {
  return (i == 0);
}

template <int d, typename position, typename ordering>
bool
heap_policy<d, position, ordering>::is_first_child (
  heap_policy<d, position, ordering>::index i
) const {
  typedef typename heap_policy<d, position, ordering>::index index;
  return (i - (i / index(d)) * index(d) == 1);
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::size () const {
  return n;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::level
heap_policy<d, position, ordering>::depth (
  heap_policy<d, position, ordering>::index i
) const {
  typedef typename heap_policy<d, position, ordering>::level level;
  typedef typename heap_policy<d, position, ordering>::index index;
  level h = 0;
  index m = 1;
  index nr = 0;
  while (i > nr) {
    m = m * index(d);
    nr = nr + m;
    ++h;
  }
  return h;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::root () const {
  return 0;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::leftmost_node (
  heap_policy<d, position, ordering>::level h

) const {
  typedef typename heap_policy<d, position, ordering>::index index;
  index r = 1;
  while (h > 0) {
    r = r * index(d);
    --h;
  }
  r = (r - 1) / index(d - 1);
  return r;
}

#ifdef BUFFER
template <int d, typename position, typename ordering>
heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::last_leaf () const {
  return l - 1;
}
#else
template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
    heap_policy<d, position, ordering>::last_leaf () const {
  return n - 1;
}
#endif

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::first_child (
  heap_policy<d, position, ordering>::index p
) const {
  return d * p + 1;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::parent (
  heap_policy<d, position, ordering>::index c
) const {
  return (c - 1) / d;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::ancestor (
  heap_policy<d, position, ordering>::index j,
  heap_policy<d, position, ordering>::level h,
  heap_policy<d, position, ordering>::level Delta
) const {
  typedef typename heap_policy<d, position, ordering>::index index;
  index t = 1;
  while (Delta > 0) {
    t = t * index(d);
    --Delta;
  }
  index i = (t - 1) / index(d - 1);
  return (j - i) / t;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::top_some_absent (
  position a,
  heap_policy<d, position, ordering>::index i,
  const ordering& less
) const {
  typedef typename heap_policy<d, position, ordering>::index index;
  typedef typename heap_policy<d, position, ordering>::element element;
  index j = heap_policy<d, position, ordering>::first_child(i);
  index k = (j + index(d) < n) ? j + index(d) : n;
  index maxc = j;
  element max = a[j];
  ++j;
  while (j < k) {
    if (less(max, a[j])) {
      maxc = j;
      max = a[j];
    }
    ++j;
  }
  return maxc;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::top_all_present (
  position a,
  heap_policy<d, position, ordering>::index i,
  const ordering& less
) const {
  return heap_policy<d, position, ordering>::top_some_absent(a, i, less);
}

template <int d, typename position, typename ordering>
void
heap_policy<d, position, ordering>::update (
  position a,
  heap_policy<d, position, ordering>::index i,
  const heap_policy<d, position, ordering>::element& x
) {
  a[i] = x;
}

#ifdef BUFFER
template <int d, typename position, typename ordering>
void
heap_policy<d, position, ordering>::erase_last_leaf (
) {
  --n;
  if (l == n) {
    if ( n <= index(d) ) {l = 1;}
    else {l = l - index(d);}
  }
}
#else
template <int d, typename position, typename ordering>
void
heap_policy<d, position, ordering>::erase_last_leaf (
) {
  --n;
}
#endif

#ifdef BUFFER
template <int d, typename position, typename ordering>
void
heap_policy<d, position, ordering>::insert_new_leaf (
) {
  ++n;
  if (( l + 2 * index(d)) == n ) {
    l = l + index(d);
  }
}
#else
template <int d, typename position, typename ordering>
void
heap_policy<d, position, ordering>::insert_new_leaf (
) {
  ++n;
}
#endif

/*

Specialization for d = 4.

*/


#ifdef BUFFER
template <> template <typename position, typename ordering>
heap_policy<4, position, ordering>::heap_policy (
  heap_policy<4, position, ordering>::index n
): n(n) {
  if(n <= 5 ) { l = 1;}
  else {
    l = ((((n - 1) >> 2) - 1) << 2) + 1;
    if ( l + 4 != n) {
      l = l + 4;
    }
  }
}
#else
template <> template <typename position, typename ordering>
heap_policy<4, position, ordering>::heap_policy (
  heap_policy<4, position, ordering>::index n
): n(n) {
}
#endif

template <> template <typename position, typename ordering>
bool
heap_policy<4, position, ordering>::is_root (
  heap_policy<4, position, ordering>::index i
) const {

  return (i == 0);
}

template <> template <typename position, typename ordering>
bool
heap_policy<4, position, ordering>::is_first_child (
  heap_policy<4, position, ordering>::index i
) const {

  return (i & 3) == 1;
}

template <> template <typename position, typename ordering>
typename heap_policy<4, position, ordering>::index
heap_policy<4, position, ordering>::size () const {

  return n;
}

template <> template <typename position, typename ordering>
typename heap_policy<4, position, ordering>::level
heap_policy<4, position, ordering>::depth (
  heap_policy<4, position, ordering>::index i
  ) const {
  typedef typename heap_policy<4, position, ordering>::level level;

  return level(ilogb((3) * i + 1) >> 1);

  /*
  // Alternativ 1
  const static index index_level[33] = {
    0, 1, 5, 21, 85, 341, 1365, 5461, 21845, 87381, 349525, 1398101, 5592405, 22369621, 89478485, 357913941, 1431655766, 1431655766, INT_MAX, INT_MAX, INT_MAX, INT_MAX, INT_MAX, INT_MAX, INT_MAX, INT_MAX, INT_MAX, INT_MAX, INT_MAX, INT_MAX, INT_MAX, INT_MAX, INT_MAX
 }; // INT_MAX added as padding

  level s = 0;

  if (index_level[16] <= i)  s = s + 16;
  if (index_level[s + 8] <= i)  s = s + 8;
  if (index_level[s + 4] <= i)  s = s + 4;
  if (index_level[s + 2] <= i)  s = s + 2;
  if (index_level[s + 1] <= i)  s = s + 1;

  return(s);
  */

  /*
  // Alternativ 2

  if (i < 1) return level(0);
  if (i < 5) return level(1);
  if (i < 21) return level(2);
  if (i < 85) return level(3);
  if (i < 341) return level(4);
  if (i < 1365) return level(5);
  if (i < 5461) return level(6);
  if (i < 21845) return level(7);
  if (i < 87381) return level(8);
  if (i < 349525) return level(9);
  if (i < 1398101) return level(10);
  if (i < 5592405) return level(11);
  if (i < 22369621) return level(12);
  if (i < 89478485) return level(13);
  if (i < 357913941) return level(14);
  if (i < 1431655766) return level(15);
  else return level(16);
  */

}

template <> template <typename position, typename ordering>
typename heap_policy<4, position, ordering>::index
heap_policy<4, position, ordering>::root () const {

  return 0;
}

template <> template <typename position, typename ordering>
typename heap_policy<4, position, ordering>::index
heap_policy<4, position, ordering>::leftmost_node (
  heap_policy<4, position, ordering>::level h
) const {

  typedef typename heap_policy<4, position, ordering>::index index;
  static const index lmn[18] = {
0, 1, 5, 21, 85, 341, 1365, 5461, 21845, 87381, 349525, 1398101, 5592405, 22369621, 89478485, 357913941, 1431655766, 1431655766
  };

  return index(lmn[h]);
}

#ifdef BUFFER
template <> template <typename position, typename ordering>
heap_policy<4, position, ordering>::index
heap_policy<4, position, ordering>::last_leaf () const {
  return l - 1;
}
#else
template <> template <typename position, typename ordering>
typename heap_policy<4, position, ordering>::index
heap_policy<4, position, ordering>::last_leaf () const {

  return n - 1;
}
#endif

template <> template <typename position, typename ordering>
typename heap_policy<4, position, ordering>::index
heap_policy<4, position, ordering>::first_child (
  heap_policy<4, position, ordering>::index p
) const {

  return (p << 2) + 1;
}

template <> template <typename position, typename ordering>
typename heap_policy<4, position, ordering>::index
heap_policy<4, position, ordering>::parent (
  heap_policy<4, position, ordering>::index c
) const {

  return (c - 1) >> 2;
}

template <> template <typename position, typename ordering>
typename heap_policy<4, position, ordering>::index
heap_policy<4, position, ordering>::ancestor (
  heap_policy<4, position, ordering>::index j,
  heap_policy<4, position, ordering>::level h,
  heap_policy<4, position, ordering>::level Delta
) const {
  typedef typename heap_policy<4, position, ordering>::index index;
  level m = 1 << (Delta << 1) ;
  level lmnd = heap_policy<4, position, ordering>::leftmost_node(h) ;
  level lmndd = heap_policy<4, position, ordering>::leftmost_node(h - Delta) ;
  return ((j - lmnd) / m) + lmndd;
}

template <> template <typename position, typename ordering>
typename heap_policy<4, position, ordering>::index
heap_policy<4, position, ordering>::top_some_absent (
  position a,
  heap_policy<4, position, ordering>::index i,
  const ordering& less
) const {

  typedef typename heap_policy<4, position, ordering>::index index;
  index j = heap_policy<4, position, ordering>::first_child(i);
  index ll = heap_policy<4, position, ordering>::size() - 1;
  index r = ll - j;
  index top;

  top = j;

  if (r == 2) {
    if (less(a[top], a[j + 1])) top = j + 1 ;
    if (less(a[top], a[j + 2])) top = j + 2 ;
    return top;
  }

  if (r == 1) {
    if (less(a[top], a[j + 1])) top = j + 1 ;
  }

  return top;
}

template <> template <typename position, typename ordering>
typename heap_policy<4, position, ordering>::index
heap_policy<4, position, ordering>::top_all_present (
  position a,
  heap_policy<4, position, ordering>::index i,
  const ordering& less
) const {
  
  typedef typename heap_policy<4, position, ordering>::index index;
  index j = heap_policy<4, position, ordering>::first_child(i);

  index top;

  top = j;

  if (less(a[top], a[j + 1])) top = j + 1 ;
  if (less(a[top], a[j + 2])) top = j + 2 ;
  if (less(a[top], a[j + 3])) top = j + 3 ;
  return top;
}

template <> template <typename position, typename ordering>
void
heap_policy<4, position, ordering>::update (
  position a,
  heap_policy<4, position, ordering>::index i,
  const heap_policy<4, position, ordering>::element& x
) {

  *(a + i) = x;
}

#ifdef BUFFER
template <> template <typename position, typename ordering>
void
heap_policy<4, position, ordering>::erase_last_leaf (
) {
  typedef typename heap_policy<4, position, ordering>::index index;

  --n;

  if(l == n) {
    if(n <= 4 ) { l = 1;}
    else {l = l - 4;}
  }

}
#else
template <> template <typename position, typename ordering>
void
heap_policy<4, position, ordering>::erase_last_leaf (
) {

  --n;
}
#endif

#ifdef BUFFER
template <> template <typename position, typename ordering>
void
heap_policy<4, position, ordering>::insert_new_leaf (
) {

  ++n;

  if (( l + 8) == n ) {
    l = l + 4;
  }
}
#else
template <> template <typename position, typename ordering>
void
heap_policy<4, position, ordering>::insert_new_leaf (
) {

  ++n;
}
#endif

#endif // __HEAP_POLICY__
