#ifndef __HEAP_POLICY__
#define __HEAP_POLICY__

#include "heap_policy.hpp" // declares heap_policy
#include <cmath> // defines std::ilogb

extern int ilogb(double) throw();

#ifdef BUFFER
template <int d, typename position, typename ordering>
heap_policy<d, position, ordering>::heap_policy (
  heap_policy<d, position, ordering>::index n
): n(n) {
  if (n <= index(d) + 1) { l = 1;}
  else {
    l = d * ((n - 1) / index(d) - 1) + 1;
    if ( l + index(d) != n) {
      l = l + index(d);
    }
  }
}
#else
template <int d, typename position, typename ordering>
heap_policy<d, position, ordering>::heap_policy (
  heap_policy<d, position, ordering>::index n
): n(n) {
}
#endif

template <int d, typename position, typename ordering>
bool
heap_policy<d, position, ordering>::is_root (
  heap_policy<d, position, ordering>::index i
) const {
  return (i == 0);
}

template <int d, typename position, typename ordering>
bool
heap_policy<d, position, ordering>::is_first_child (
  heap_policy<d, position, ordering>::index i
) const {
  typedef typename heap_policy<d, position, ordering>::index index;
  return (i - (i / index(d)) * index(d) == 1);
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::size () const {
  return n;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::level
heap_policy<d, position, ordering>::depth (
  heap_policy<d, position, ordering>::index i
) const {
  typedef typename heap_policy<d, position, ordering>::level level;
  typedef typename heap_policy<d, position, ordering>::index index;
  level h = 0;
  index m = 1;
  index nr = 0;
  while (i > nr) {
    m = m * index(d);
    nr = nr + m;
    ++h;
  }
  return h;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::root () const {
  return 0;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::leftmost_node (
  heap_policy<d, position, ordering>::level h
) const {
  typedef typename heap_policy<d, position, ordering>::index index;
  index r = 1;
  while (h > 0) {
    r = r * index(d);
    --h;
  }
  r = (r - 1) / index(d - 1);
  return r;
}

#ifdef BUFFER
template <int d, typename position, typename ordering>
heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::last_leaf () const {
  return l - 1;
}
#else
template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::last_leaf () const {
  return n - 1;
}
#endif

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::first_child (
  heap_policy<d, position, ordering>::index p
) const {
  return d * p + 1;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::parent (
  heap_policy<d, position, ordering>::index c
) const {
  return (c - 1) / d;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::ancestor (
  heap_policy<d, position, ordering>::index j,
  heap_policy<d, position, ordering>::level h,
  heap_policy<d, position, ordering>::level Delta
) const {
  typedef typename heap_policy<d, position, ordering>::index index;
  index t = 1;
  while (Delta > 0) {
    t = t * index(d);
    --Delta;
  }
  index i = (t - 1) / index(d - 1);
  return (j - i) / t;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::top_some_absent (
  position a,
  heap_policy<d, position, ordering>::index i,
  const ordering& less
) const {
  typedef typename heap_policy<d, position, ordering>::index index;
  typedef typename heap_policy<d, position, ordering>::element element;
  index j = heap_policy<d, position, ordering>::first_child(i);
  index k = (j + index(d) < n) ? j + index(d) : n;
  index maxc = j;
  element max = a[j];
  ++j;
  while (j < k) {
    if (less(max, a[j])) {
      maxc = j;
      max = a[j];
    }
    ++j;
  }
  return maxc;
}

template <int d, typename position, typename ordering>
typename heap_policy<d, position, ordering>::index
heap_policy<d, position, ordering>::top_all_present (
  position a,
  heap_policy<d, position, ordering>::index i,
  const ordering& less
) const {
  return heap_policy<d, position, ordering>::top_some_absent(a, i, less);
}

template <int d, typename position, typename ordering>
void
heap_policy<d, position, ordering>::update (
  position a,
  heap_policy<d, position, ordering>::index i,
  const heap_policy<d, position, ordering>::element& x
) {
  a[i] = x;
}

#ifdef BUFFER
template <int d, typename position, typename ordering>
void
heap_policy<d, position, ordering>::erase_last_leaf (
) {
  --n;
  if (l == n) {
    if ( n <= index(d) ) {l = 1;}
    else {l = l - index(d);}
  }
}
#else
template <int d, typename position, typename ordering>
void
heap_policy<d, position, ordering>::erase_last_leaf (
) {
  --n;
}
#endif

#ifdef BUFFER
template <int d, typename position, typename ordering>
void
heap_policy<d, position, ordering>::insert_new_leaf (
) {
  ++n;
  if (( l + 2 * index(d)) == n ) {
    l = l + index(d);
  }
}
#else
template <int d, typename position, typename ordering>
void
heap_policy<d, position, ordering>::insert_new_leaf (
) {
  ++n;
}
#endif


/*

Specialization for d = 8.

*/

#ifdef BUFFER
template <> template <typename position, typename ordering>
heap_policy<8, position, ordering>::heap_policy (
  heap_policy<8, position, ordering>::index n
): n(n) {
  if (n <= 9) { l = 1;}
  else {
    l = 8 * ((n - 1) / 8 - 1) + 1;
    if ( l + 8 != n) {
      l = l + 8;
    }
  }
}
#else
template <> template <typename position, typename ordering>
heap_policy<8, position, ordering>::heap_policy (
  heap_policy<8, position, ordering>::index n
): n(n) {
}
#endif

template <> template <typename position, typename ordering>
bool
heap_policy<8, position, ordering>::is_root (
  heap_policy<8, position, ordering>::index i
) const {

  return (i == 0);
}

template <> template <typename position, typename ordering>
bool
heap_policy<8, position, ordering>::is_first_child (
  heap_policy<8, position, ordering>::index i
) const {

  return (i & 7) == 1;
}

template <> template <typename position, typename ordering>
typename heap_policy<8, position, ordering>::index
heap_policy<8, position, ordering>::size () const {

  return n;
}

template <> template <typename position, typename ordering>
typename heap_policy<8, position, ordering>::level
heap_policy<8, position, ordering>::depth (
  heap_policy<8, position, ordering>::index i
  ) const {
  typedef typename heap_policy<8, position, ordering>::level level;

  const static level div3[] = {
    0,0,0,1,1,1,2,2,2,3,3,3,4,4,4,5,5,5,6,6,6,7,7,7,8,8,8,9,9,9,
    10,10,10,11,11,11,12,12,12,13,13,13,14,14,14,15,15,15,16,16,16,17,17,17,18,18,18,19,19,19,
    20,20,20,21,21,21,22,22,22,23,23,23,24,24,24,25,25,25,26,26,26,27,27,27,28,28,28,29,29,29,
    30,30,30,31,31,31,32,32,32,33,33,33,34,34,34,35,35,35,36,36,36,37,37,37,38,38,38,38,38,39,39,39,
    40,40,40,41,41,41,42,42,42
  };

  if (i == 0) return 0;
  if (i < 9) return 1;
  if (i < 18) return 2;
  return level(div3[ilogb(7 * i + 1)]);
}

template <> template <typename position, typename ordering>
typename heap_policy<8, position, ordering>::index
heap_policy<8, position, ordering>::root () const {

  return 0;
}

template <> template <typename position, typename ordering>
typename heap_policy<8, position, ordering>::index
heap_policy<8, position, ordering>::leftmost_node (
  heap_policy<8, position, ordering>::level h
) const {
  typedef typename heap_policy<8, position, ordering>::index index;
  const static index lmn[12] = {
    0, 1, 9, 73, 585, 4681, 37449, 299593, 2396745, 19173961, 153391689, 1227133513
  };
 return lmn[h];
}

#ifdef BUFFER
template <> template <typename position, typename ordering>
heap_policy<8, position, ordering>::index
heap_policy<8, position, ordering>::last_leaf () const {
  return l - 1;
}
#else
template <> template <typename position, typename ordering>
typename heap_policy<8, position, ordering>::index
heap_policy<8, position, ordering>::last_leaf () const {
  return n - 1;
}
#endif

template <> template <typename position, typename ordering>
typename heap_policy<8, position, ordering>::index
heap_policy<8, position, ordering>::first_child (
  heap_policy<8, position, ordering>::index p
) const {

  return (p << 3) + 1;
}

template <> template <typename position, typename ordering>
typename heap_policy<8, position, ordering>::index
heap_policy<8, position, ordering>::parent (
  heap_policy<8, position, ordering>::index c
) const {

  return (c - 1) >> 3;
}

template <> template <typename position, typename ordering>
typename heap_policy<8, position, ordering>::index
heap_policy<8, position, ordering>::ancestor (
  heap_policy<8, position, ordering>::index j,
  heap_policy<8, position, ordering>::level h,
  heap_policy<8, position, ordering>::level Delta
) const {
  typedef typename heap_policy<8, position, ordering>::index index;

  if (Delta == 0) return j;
  index m = 8 << (Delta - 1) * 3;
  level lmnh = heap_policy<8, position, ordering>::leftmost_node(h) ;
  level lmnhd = heap_policy<8, position, ordering>::leftmost_node(h - Delta) ;
  return ((j - lmnh) / m) + lmnhd;

}

template <> template <typename position, typename ordering>
typename heap_policy<8, position, ordering>::index
heap_policy<8, position, ordering>::top_some_absent (
  position a,
  heap_policy<8, position, ordering>::index i,
  const ordering& less
) const {
  typedef typename heap_policy<8, position, ordering>::index index;
  index j = heap_policy<8, position, ordering>::first_child(i);
  index ll = heap_policy<8, position, ordering>::size() - 1;
  index r = ll - j;
  index top;

  top = j;

  if (r == 6) {
  if (less(a[top], a[j + 1])) top = j + 1 ;
  if (less(a[top], a[j + 2])) top = j + 2 ;
  if (less(a[top], a[j + 3])) top = j + 3 ;
  if (less(a[top], a[j + 4])) top = j + 4 ;
  if (less(a[top], a[j + 5])) top = j + 5 ;
  if (less(a[top], a[j + 6])) top = j + 6 ;
  return top;
  }

  if (r == 5) {
  if (less(a[top], a[j + 1])) top = j + 1 ;
  if (less(a[top], a[j + 2])) top = j + 2 ;
  if (less(a[top], a[j + 3])) top = j + 3 ;
  if (less(a[top], a[j + 4])) top = j + 4 ;
  if (less(a[top], a[j + 5])) top = j + 5 ;
  return top;
  }

  if (r == 4) {
  if (less(a[top], a[j + 1])) top = j + 1 ;
  if (less(a[top], a[j + 2])) top = j + 2 ;
  if (less(a[top], a[j + 3])) top = j + 3 ;
  if (less(a[top], a[j + 4])) top = j + 4 ;
  return top;
  }

  if (r == 3) {
  if (less(a[top], a[j + 1])) top = j + 1 ;
  if (less(a[top], a[j + 2])) top = j + 2 ;
  if (less(a[top], a[j + 3])) top = j + 3 ;
  return top;
  }

  if (r == 2) {
  if (less(a[top], a[j + 1])) top = j + 1 ;
  if (less(a[top], a[j + 2])) top = j + 2 ;
  return top;
  }

  if (r == 1) {
  if (less(a[top], a[j + 1])) top = j + 1 ;
  return top;
  }

  return top;
}

template <> template <typename position, typename ordering>
typename heap_policy<8, position, ordering>::index
heap_policy<8, position, ordering>::top_all_present (
  position a,
  heap_policy<8, position, ordering>::index i,
  const ordering& less
) const {
  typedef typename heap_policy<8, position, ordering>::index index;
  index j = heap_policy<8, position, ordering>::first_child(i);
  index top;

  top = j;

  if (less(a[top], a[j + 1])) top = j + 1 ;
  if (less(a[top], a[j + 2])) top = j + 2 ;
  if (less(a[top], a[j + 3])) top = j + 3 ;
  if (less(a[top], a[j + 4])) top = j + 4 ;
  if (less(a[top], a[j + 5])) top = j + 5 ;
  if (less(a[top], a[j + 6])) top = j + 6 ;
  if (less(a[top], a[j + 7])) top = j + 7 ;
  return top;
}

template <> template <typename position, typename ordering>
void
heap_policy<8, position, ordering>::update (
  position a,
  heap_policy<8, position, ordering>::index i,
  const heap_policy<8, position, ordering>::element& x
) {
  a[i] = x;
}

#ifdef BUFFER
template <> template <typename position, typename ordering>
void
heap_policy<8, position, ordering>::erase_last_leaf (
) {
  --n;
  if (l == n) {
    if ( n <= 8 ) {l = 1;}
    else {l = l - 8;}
  }
}
#else
template <> template <typename position, typename ordering>
void
heap_policy<8, position, ordering>::erase_last_leaf (
) {
  --n;
}
#endif

#ifdef BUFFER
template <> template <typename position, typename ordering>
void
heap_policy<8, position, ordering>::insert_new_leaf (
) {
  ++n;
  if (( l + 2 * 8) == n ) {
    l = l + 8;
  }
}
#else
template <> template <typename position, typename ordering>
void
heap_policy<8, position, ordering>::insert_new_leaf (
) {

  ++n;
}
#endif

#endif // __HEAP_POLICY__
