#include <iostream>
#include <functional>
#include <cassert>

#include "heap_policy.cpp"

int main() {

  typedef int integer;
  integer n = 4;
  typedef int element;
  typedef element* position;
  typedef std::less<element> ordering;
  heap_policy<2, position, ordering> heap(n);
  ordering less = ordering();
  int a[4] = {0, 1, 2, 3};

  std::cout << 
    "testing is_root " << std::endl;
  assert(heap.is_root(0) == true);
  assert(heap.is_root(1) == false);
  assert(heap.is_root(2) == false);

  std::cout << 
    "testing is_first_child " << std::endl;
  assert(heap.is_first_child(0) == false);
  assert(heap.is_first_child(1) == true);
  assert(heap.is_first_child(2) == false);
  assert(heap.is_first_child(3) == true);

  std::cout << 
    "testing root " << std::endl;
  assert(heap.root() == 0);

  std::cout << 
    "testing size " << std::endl;
  assert(heap.size() == n);

  std::cout << 
    "testing depth " << std::endl;
  assert(heap.depth(0) == 0);
  assert(heap.depth(1) == 1);
  assert(heap.depth(2) == 1);
  assert(heap.depth(3) == 2);

  std::cout << 
    "testing leftmost_node "  << std::endl;
  assert(heap.leftmost_node(2) == 3);

  std::cout << 
    "testing last_leaf "  << std::endl;
  assert(heap.last_leaf() == n - 1);
  assert(heap.last_leaf() == heap.size() - 1);

  std::cout << 
    "testing first_child "  << std::endl;
  assert(heap.first_child(0) == 1);
  assert(heap.first_child(1) == 3);
  assert(heap.first_child(2) == 5);
  assert(heap.first_child(3) == 7);

  std::cout << 
    "testing parent " << std::endl;
  assert(heap.parent(1) == 0);
  assert(heap.parent(2) == 0);
  assert(heap.parent(3) == 1);
  assert(heap.parent(4) == 1);
  assert(heap.parent(5) == 2);
  assert(heap.parent(6) == 2);

  std::cout << 
    "testing ancestor " << std::endl;
  assert(heap.ancestor(0, 0, 0) == 0);
  assert(heap.ancestor(1, 1, 0) == 1);
  assert(heap.ancestor(1, 1, 1) == 0);
  assert(heap.ancestor(2, 1, 1) == 0);
  assert(heap.ancestor(3, 2, 0) == 3);
  assert(heap.ancestor(3, 2, 1) == 1);
  assert(heap.ancestor(3, 2, 2) == 0);

  std::cout << 
    "testing top_some_absent " << std::endl;
  assert(heap.top_some_absent(a, 0, less) == 2);
  assert(heap.top_some_absent(a, 1, less) == 3);

  std::cout << 
    "testing top_all_present " << std::endl;
  assert(heap.top_all_present(a, 0, less) == 2);

  std::cout << 
    "testing erase_last_leaf " << std::endl;
  heap.erase_last_leaf();
  assert(heap.size() == n - 1);

  std::cout << 
    "testing insert_new_leaf " << std::endl;
  heap.insert_new_leaf();
  assert(heap.size() == n);
  heap.insert_new_leaf();
  assert(heap.leftmost_node(heap.depth(heap.last_leaf())) == 3);

  return 0;
}
