"""
Benchmarking various versions of if statments

"""
import benchmark
import os

class if_case(benchmark.case):
  def __init__(self, n, element, function, directory):
    benchmark.case.__init__(self)
    self.n = n
#    self.computer = 'pc-014.diku.dk'
#    self.compiler = '/scratch/g++3.0.4/bin/g++'
    self.compiler = '/usr/bin/g++-3.2.2' 
    self.compiler_options.extend(['-O6'])
    home = os.environ['HOME']
    self.include_paths.extend([
      home + '/CPHSTL/Report/In-place-multiway-heaps/Program/Microbenchmarks/' + directory + '/' 
      ])
    self.include_files.extend(['<vector>', '<functional>', 'registers.c++', function + '.c++'])
    self.dual_exists = 0
    self.constructor_call = \
      'experiment(' + str(n) + ')'
    self.time_unit = 'ns'
    self.driver_file = self.generate_cpu_time_driver()

  def output(self):
    return (self.n, float(self.driver_output) / self.n)
 # def tidy_up(self): 
 #   "Used for debugging to see the files generated"
 #   pass 

class if_curve(benchmark.curve_suite):
  def __init__(self, element, experiment, functions, log_i, log_k):
    benchmark.curve_suite.__init__(self)
    self.title = 'experiment: ' + experiment + ' source: ' + functions 
    for j in range(log_i, log_k + 1):
      self.add(if_case(1 << j, element, experiment, functions))

class if_plot(benchmark.plot_suite):
  def __init__(self, element, log_i, log_k):
    benchmark.plot_suite.__init__(self)
    self.title = 'Efficiency of various use of registers integers' 
    self.xlabel = 'n'
    self.ylabel = 'Execution time per n [in nanoseconds]'
    self.add(if_curve(element, 'cost_reg', 'Registers3', log_i, log_k))
    self.add(if_curve(element, 'cost_reg', 'Registers8', log_i, log_k))
    self.add(if_curve(element, 'cost_reg', 'Registers16', log_i, log_k))
    self.add(if_curve(element, 'cost_reg', 'Registers16EI', log_i, log_k))
    self.gnuplot_commands += """
set title '%(title)s'
set xlabel '%(xlabel)s'
set ylabel '%(ylabel)s'
set logscale x
""" % self.__dict__

if __name__ == '__main__':
  benchmark.main(
    task = if_plot('int', 5, 15),
    runner = benchmark.gnuplot_runner
  )
