#ifndef __HEAP_FUNCTIONS__
#define __HEAP_FUNCTIONS__

template <int d, typename position, typename ordering, typename policy>
void make_heap(position, const ordering&, policy&);

#ifndef __HEAP_POLICY__
#include "heap_policy.cpp" // defines the heap policy
#endif
#ifndef __UTILITY_FUNCTIONS__
#include "utility_functions.cpp" // defines the functions needed by heap functions
#endif

#include "debugging.c++"

template <int d, typename position, typename ordering, typename policy>
void pop_heap(position a, const ordering& less, policy& p) {
  //std::cout << "before pop_heap" << std::endl;
  //show(a, a + p.size());
 
  typedef typename policy::index index;
  typedef typename policy::element element;

  if (p.is_root(p.size() - 1)) {
    return;
  }

  index i = p.root();
  index last_buffer_leaf = p.size() - 1;
  index buffer_parent = p.parent(last_buffer_leaf);
  index buffer_size = last_buffer_leaf - p.last_leaf();

  index j, s ,t;
  element x;
  
  if (buffer_size < d){
    j = p.top_some_absent(a, buffer_parent, less);
    if (less(a[i], a[j])) {
      if (j != last_buffer_leaf) {
	x = a[j];
	p.update(a, j, a[last_buffer_leaf]);
	a[last_buffer_leaf] = x;
      }
      p.erase_last_leaf();
    }
    else {
      x = a[j];
      if (j != last_buffer_leaf) {
	p.update(a, j, a[last_buffer_leaf]);
      }
      a[last_buffer_leaf] = a[i];
      p.erase_last_leaf();
      sift_down<d, position, index, element, ordering, policy>(a, i, x, less, p);
    }
    return;
  }

  if (buffer_size > d) {
    s = p.top_all_present(a, buffer_parent - 1, less);
    t = p.top_some_absent(a, buffer_parent, less);
    if (less(a[t], a[s])) {
      x = a[s];
      p.update(a, s, a[last_buffer_leaf]);
      if (less(a[s], a[i])) {
	a[last_buffer_leaf] = a[i];
	p.erase_last_leaf();
	sift_down<d, position, index, element, ordering, policy>(a, i, x, less, p);
      }
      else {
	a[last_buffer_leaf] = x;
	p.erase_last_leaf();
      }
    }
    else {
      if (less(a[t], a[i])) {
	x = a[t];
	if (t != last_buffer_leaf) {
	  p.update(a, t, a[last_buffer_leaf]);
	} 
	a[last_buffer_leaf] = a[i];
	p.erase_last_leaf();
	sift_down<d, position, index, element, ordering, policy>(a, i, x, less, p);
      }
      else {
	if (t != last_buffer_leaf) {
	  x = a[t];
	  p.update(a, t, a[last_buffer_leaf]);
	  a[last_buffer_leaf] = x;
	}
	p.erase_last_leaf();
      }
    }
    return;
  }

  if (buffer_size == d){
    s = p.top_all_present(a, buffer_parent, less);
    if (less(a[s], a[i])) {
      x = a[s];
      if (s != last_buffer_leaf) {
	p.update(a, s, a[last_buffer_leaf]);
      }
      a[last_buffer_leaf] = a[i];
      p.erase_last_leaf();
      sift_down<d, position, index, element, ordering, policy>(a, i, x, less, p);
    }
    else {
      if (s != last_buffer_leaf) {
	x = a[s];
	p.update(a, s, a[last_buffer_leaf]);
	a[last_buffer_leaf] = x;
      }
      p.erase_last_leaf();
    }
  }

  //std::cout << "after pop_heap" << std::endl;
  //show(a, a + p.size());
 }


template <int d, typename position, typename ordering, typename policy>
void push_heap(position a, const ordering& less, policy& p) {
  //std::cout << "before push_heap" << std::endl;
  //show(a, a + p.size());

  typedef typename policy::index index;
  typedef typename policy::element element;

  index last_buffer_leaf = p.size() - 1;

  if ((p.last_leaf() + 2 * d - 1) == last_buffer_leaf ) {
    sift_up_buffer<d, position, index, element, ordering, policy>(a, less, p);
  }

  p.insert_new_leaf();

  last_buffer_leaf = p.size() - 1;

  p.update(a, last_buffer_leaf, a[last_buffer_leaf]);

  //std::cout << "after push_heap" << std::endl;
  //show(a, a + p.size());
 }



template <int d, typename position, typename ordering, typename policy>
void make_heap(position a, const ordering& less, policy& p) {
  if (p.size() < 2) return;
  //std::cout << "before make_heap: ";
  //show(a, a + p.size());

  typedef typename policy::index index;
  typedef typename policy::element element;
  const index m = p.leftmost_node(p.depth(p.size() - 1));
  const index j = p.parent(m);
  const index k = p.parent(p.size() - 1);

  for (index l = m - 1; l >= j; --l) {
    if (l <= k) {
      sift_down<d, position, index, element, ordering, policy>(a, l, *(a + l), less, p);
    }
    for (index i = l; p.is_first_child(i); ) {
      i = p.parent(i);
      sift_down<d, position, index, element, ordering, policy>(a, i, *(a + i), less, p);
    }
  }
  //std::cout << "after make_heap: ";
  //show(a, a + p.size());
}

template <int d, typename position, typename ordering, typename policy>
void sort_heap(position a, const ordering& less, policy& p) {
  while (p.size() > 1) {
    pop_heap<d, position, ordering, policy>(a, less, p);
  }
}


#include <is_sorted.c++> // defines cphstl::is_sorted

template <int d, typename position, typename ordering, typename policy>
bool is_heap(position a, position b, const ordering& less, policy& p) {
  if (p.size() < 1) return true;
  typedef typename heap_policy<d, position, ordering>::index index;

  index last_branch = p.parent(p.last_leaf());
  for (index i = p.root(); i <= last_branch; ++i) {
    index j = p.first_child(i);
    index k = j + index(d) - 1;
    k = (k <= p.last_leaf()) ? k : p.last_leaf();
    while (j <= k) {
      if (less(*(a + i), *(a + j))) return false;
      ++j;
    }
  }
  return true;
}

#include <algorithm>
#include <vector>

template <int d, typename position, typename ordering>
void heapsort(position a, position b, const ordering& less) {
  typedef heap_policy<d, position, ordering> policy;
  policy p(b - a);

  typedef typename policy::element element;
  std::vector<element> t(b - a);
  std::copy(a, b, t.begin());
  std::sort(t.begin(), t.end(), ordering());

  //std::cout << "before make_heap: ";
  //show(a, b);

  make_heap<d, position, ordering, policy>(a, less, p);
  bool result = is_heap<d, position, ordering, policy>(a, b, less, p);
  assert(result);

  //std::cout << "before sort_heap: ";
  //show(a, b);

  sort_heap<d, position, ordering, policy>(a, less, p);

  //std::cout << "after sort_heap: ";
  //show(a, b);

  //std::cout << "STL sort: ";
  //show(t.begin(), t.end());

  assert(cphstl::is_sorted(a, b, less));
}


template <int d, typename position, typename ordering, typename policy>
void sort(position a, position b, const ordering& less) {
  policy p(b - a);

  make_heap<d, position, ordering, policy>(a, less, p);
  sort_heap<d, position, ordering, policy>(a, less, p);
}

#endif // __HEAP_FUNCTIONS__
