#ifndef __UTILITY_FUNCTIONS__
#define __UTILITY_FUNCTIONS__

#include <iostream>

template <int d, typename position, typename index, typename element, 
  typename ordering, typename policy>
void
sift_down(position a, index i, element x, const ordering& less, policy& p) {
  //std::cout << "before sift_down i,x: " << i << " " << x << std::endl;
  //show(a, a + p.size());

  index m = i;  
  index j = p.first_child(i);
  const index last_full_branch = p.parent(p.last_leaf() + 1) - 1;
  
  while (i <= last_full_branch) {
    j = p.top_all_present(a, i, less);
    p.update(a, i, a[j]);
    i = j;
  }
  
  j = p.first_child(i);
  if (j < p.size()) {
    j = p.top_some_absent(a, i, less);      
    p.update(a, i, a[j]);
    if (less(a[i], x)) {
      p.update(a, j, a[i]);
    }
    else {
      p.update(a, j, x);
      return;
    }
  }

  index k;
  
  if ((i == m) || !less(a[p.parent(i)], x)) {
    p.update(a, i, x);
    return;
  }
  
  while (true) {
    k = i;
    j = p.parent(i);
    i = p.parent(j);
    
    if ((i > m) && less(a[i], x)) {
      p.update(a, k, a[j]);
      if (i == m) break;
      p.update(a, j, a[i]);
    }
    else break;
  }

  if (i == m) {
    if (less(a[i], x)) {
      p.update(a, k, a[j]);
      p.update(a, j, a[i]);
      p.update(a, i, x);
      return;
    }
  }

  if (less(a[j], x)) {
    p.update(a, k, a[j]);
    p.update(a, j, x);
    return;
  }
  else {
    p.update(a, k, x);
    return;
  }
  //std::cout << "after sift_down: " << std::endl;
  //show(a, a + p.size());
} 

template <int d, typename position, typename index, typename element, 
  typename ordering, typename policy>
void
sift_up(position a, index j, element x, const ordering& less, policy& p) {
  //std::cout << "before sift_up j: " << j << std::endl;
  //show(a, a + p.size());

  index i;
  index k;

  i = j;
  if ((i == p.root()) || !less(a[p.parent(i)], x)) {
    p.update(a, i, x);
    return;
  }
  
  while (true) {
    k = i;
    j = p.parent(i);
    i = p.parent(j);
    
    if ((i > p.root()) && less(a[i], x)) {
      p.update(a, k, a[j]);
      if (i == p.root()) break;
      p.update(a, j, a[i]);
    }
    else break;
  }

  if (i == p.root()) {
    if (less(a[i], x)) {
      p.update(a, k, a[j]);
      p.update(a, j, a[i]);
      p.update(a, i, x);
      return;
    }
  }

  if (less(a[j], x)) {
    p.update(a, k, a[j]);
    p.update(a, j, x);
    return;
  }
  else {
    p.update(a, k, x);
    return;
  }

  //std::cout << "after sift_up" << std::endl;
  //show(a, a + p.size());
} 

#endif // __UTILITY_FUNCTIONS__


