/*
  This is an implementation of the in-place merge algorithm
  of mannila and ukkonen in this first version the algorithm
  is actually not in place but uses O(sqrt(n)) extra space.
  We will strive to remove this workspace in the next version.  
*/
#include <math.h>
#include <stdlib.h>
#include <iterator>


namespace cphstl {
 
 //copymerge er den specielle merge vi har brug for for at kunne
 //benytte interne buffere, den almindelige merge overskriver jo
 //output bufferen. first2 skal vre lig med result+first1-last1+1
 template <typename ForwardIterator,class T, class comparator>
 void copymerge(ForwardIterator firstAtemp, ForwardIterator lastAtemp,
                  ForwardIterator firstB, ForwardIterator lastB,
                  ForwardIterator firstA, T tempT, comparator comp) {
  while (firstAtemp!=lastAtemp && firstB!=lastB){
    if(comp(*firstAtemp, *firstB)<0){
      tempT=*firstA;
      *firstA=*firstAtemp;   
      *firstAtemp=tempT;
      firstAtemp++;
    }else {
      tempT=*firstA;
      *firstA=*firstB;
      *firstB=tempT;
      firstB++;
    }
    firstA++;
  }

  if (firstB==lastB) {
    while(firstAtemp!=lastAtemp){
      tempT=*firstA;
      *firstA=*firstAtemp;
      *firstAtemp=tempT;
      firstA++;
      firstAtemp++;
    }
     
  } 
  return;
 } 
 
 template <typename ForwardIterator, typename Size, class T>
 void swap_n(ForwardIterator list1, ForwardIterator list2,
 	Size n, T tempT) {
   while(n>0) {
     tempT=*list1;
     *list1=*list2;
     *list2=tempT;
     list2++;
     list1++;
     n--;
   }
   return ;
 }
//use the cphstl version instead when it is implemented
template <typename ForwardIter, class Tp, class comparator>
 ForwardIter upper_bound_ukko(ForwardIter first, ForwardIter last,
                              const Tp& val, comparator comp)
 {
   int len = 0;
   ForwardIter temp=first;
   len = std::distance(temp,last);
      
   int half;
   int i;
   ForwardIter middle;
      
   while (len > 0) {
     half = len >> 1;
     middle = first;
     i=half;
     std::advance(middle,i);
     if (comp(val,*middle)<0){
       len = half;
     }else {
       first = middle;  
       first++;
       len = len - half - 1;
     }
   }
   return first;
 }
 
 //denne funktion implementerer den interchange der beskrives i artiklen
 template<typename  BII, class T>
  void interchange(BII first, BII middle, BII last, T tempT){
     BII temp, temp2;
     
     temp=first;
     temp2=middle;
     temp2--;
     while(temp<temp2){
       tempT=*temp;
       *temp=*temp2;
       *temp2=tempT;
       temp++;
       temp2--;
     }
     temp=middle;
     temp2=last;
     temp2--;
     while(temp<temp2){
       tempT=*temp;
       *temp=*temp2;
       *temp2=tempT;
       temp++;
       temp2--;
     }
     temp=first;
     temp2=last;
     temp2--;
     while(temp<temp2){
       tempT=*temp;
       *temp=*temp2;
       *temp2=tempT;
       temp++;
       temp2--;
     }
     return;
  }
 
 template <typename FWDI,class T, class comparator>
 void insertionsort(FWDI first, FWDI last, T tempT, comparator comp){
   FWDI temp,temp2;
   while (first!=last){
     temp2=temp=first;
     temp2++;
     while(temp2!=last){
       if (comp(*temp2,*temp)<0) temp=temp2;
       temp2++;
     }
     if(first!=temp){
           tempT=*first;
           *first=*temp;
           *temp=tempT;
     }
     first++;
   }
   return;
 }
 
 //in this new version we move the short block down first 
 template<typename BDiterator, typename comparator>
 void inplace_merge_ukkonen(BDiterator A, BDiterator B,BDiterator end, comparator comp){
   using namespace std;
   int len1=0;
   int len2=0;
   
   len1 = std::distance(A,B);
   len2 = std::distance(B,end);
      
   if ((len1 == 0) || (len2 == 0)) return;
   
   BDiterator temp,temp2,temp3;
    
   //if the A array does not contain at least four blocks the final merging step will
   //fail so we must make sure there are at least four blocks. if len1 is at least
   // 20 then there must be four blocks so from 10 and up we can use this algorithm
   typedef typename std::iterator_traits<BDiterator>::value_type T;
   T tempT;
   
   if (len1 < 20) {
     //here we should do an selection sort or something faster.
     temp=A;
     for(int k =0 ;k<(len1+len2); k++){
         temp2=temp;
         temp3=temp;
         for (int l=k;l<(len1+len2);l++){
           if (comp(*temp3,*temp2)<0) temp2=temp3;
           temp3++;
         }
         if(temp!=temp2){
           tempT=*temp;
           *temp=*temp2;
           *temp2=tempT;
         }
         temp++;
       }  
      return;
   }
    
   //rearrange the array so that it is blockwise sorted i.e. a1.a2..b1.b2.. -> a1.b1.a2.b2..
   BDiterator ms=A; //midsection start pointer
   BDiterator ls=B; //last section start pointer
   BDiterator ajl,ajf,oldms;
   int blocksize =(int) ceil(sqrt((double) len1));
   int lastblocksize =len1-((blocksize - ((blocksize*blocksize)-len1) / blocksize-1)*blocksize);
  
  
   int h;
   int splitamount,aksplit;
   int compint;  
   
   h=0;
   splitamount=0;
  
   while (ms<ls) {
     oldms=ms;
         
     //find next A block to be moved
     aksplit=0;
     if (ms!=A) {
       temp=ms;
       std::advance(temp,blocksize-splitamount-1);
       if (splitamount) aksplit=1;//this is the split block  
       ajl=temp;
       std::advance(temp,blocksize);
       while (temp<ls) {
         compint =comp(*temp,*ajl); 
         if (compint<0) {
	   ajl=temp;
	   aksplit=0; //we have chosen a block that is not split 
	 }
	 if (compint==0){
	   if (aksplit){
	     temp2 = ls;
	     std::advance(temp2,-splitamount);
	   }else{
	     temp2 = ajl;
	     std::advance(temp2,1 - blocksize);
	   }
	   temp3 = temp;
	   std::advance(temp3, 1 - blocksize); 
	   compint = comp(*temp3,*temp2);
	   if (compint<0) {
	     ajl=temp;
	     aksplit=0; //we have chosen a block that is not split 
	   } 
	 }
         std::advance(temp,blocksize); 
       }
       ajf=ajl;
       std::advance(ajf, 1 - blocksize);
     } else {
       aksplit=0;
       ajf= ajl= ms;
       std::advance(ajl,lastblocksize - 1);
     }
     
     //move the block
     if (aksplit) { //we know it is the first block in C that is split
       //first we exchange the split part 
       ajf=ls;
       std::advance(ajf,-splitamount);
       temp=ajf;
       temp2=ajl;
       temp2++;
       if (temp!=temp2){ //else this is the last block and we do not have to do this
         for (int i =0; i<splitamount; i++){
	   tempT=*temp;
           *temp=*temp2;
           *temp2=tempT;
	   temp++;
	   temp2++;
	 }
       }
        //now we rotate the block into the correct order
       temp2=ajl;
       temp2++;
       temp3=ms;
       std::advance(ms,blocksize);
       interchange(temp3,temp2,ms,tempT);
       
     } else if (ms==ajf){ // the block is first in the middle area C
       if (ms==A) std::advance(ms,lastblocksize);
       else std::advance(ms,blocksize); 
     } else {
       //if ajf is in the block just after the splitblock we have to do something else
       temp=ms;
       std::advance(temp,blocksize-splitamount);
       if (temp==ajf) {
         temp2=ajl;
	 temp2++;
	 interchange(ms,ajf,temp2,tempT);
	 std::advance(ms,blocksize);
       }else{
         temp=ajf;
         for(h=0;h<blocksize;h++,ms++,temp++){
           tempT=*temp;
           *temp=*ms;
           *ms=tempT;
         }
         temp=ajf;
	 std::advance(temp,blocksize-splitamount);
         temp2=ls;
	 std::advance(temp2,-splitamount);
         for(h=0;h<splitamount;h++,temp++,temp2++){
           tempT=*temp;
           *temp=*temp2;
           *temp2=tempT;
         }
	 temp=ajf;
	 std::advance(temp,blocksize-splitamount);
	 temp2=ajl;
	 temp2++;
         interchange(ajf,temp,temp2,tempT);
       }
     }
     
     //move the Bi elements down in front
     
     temp2=ms;
     temp2--;
     if (ms<ls) temp=upper_bound_ukko(ls,end,*temp2,comp);
     else temp=end;
     
               
     while (ls<temp){
       tempT=*ls;
       *ls=*ms;
       *ms=tempT;
       
       ls++;
       ms++;
       splitamount++;
       if (splitamount==blocksize) splitamount=0;
     }
    
     //now we have to merge
    
     if (oldms==A) {//happens once
       temp=ls;
       std::advance(temp,-lastblocksize-splitamount);
       temp2=temp;
       std::advance(temp2,lastblocksize);
       temp3=A;
       std::advance(temp3,lastblocksize);
       
       swap_n(A,temp,lastblocksize,tempT);
       copymerge(temp,temp2,temp3,ms,A,tempT,comp);
             
       insertionsort(temp,temp2,tempT,comp); 
     }else if ((oldms-A)>=blocksize){
       temp2=A;
       std::advance(temp2,blocksize);
       temp3=oldms;
       std::advance(temp3,blocksize);
       
       swap_n(oldms,A,blocksize,tempT);
       copymerge(A,temp2,temp3,ms,oldms,tempT,comp);
     }else{//happens at most once
       //we assume that there are at least four A blocks
       temp=ls;
       std::advance(temp,-blocksize-splitamount);
       temp2=temp;
       std::advance(temp2,blocksize);
       //then these pointers are pointing to an unsplit block
       temp3=oldms;
       std::advance(temp3,blocksize);
       
       swap_n(oldms,temp,blocksize,tempT);
       copymerge(temp,temp2,temp3,ms,oldms,tempT,comp);
      
       insertionsort(temp,temp2,tempT,comp);      
     }
   }
    
   //now we have to sort our temporary buffer:
   temp=A;
   std::advance(temp,blocksize);
   insertionsort(A,temp,tempT,comp);
   return;
 }

}
